# Copyright 2017-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class AutounmaskUseSlotConflictTestCase(TestCase):
    def testAutounmaskUseSlotConflict(self):
        self.todo = True

        ebuilds = {
            "sci-libs/K-1": {"IUSE": "+foo", "EAPI": 1},
            "sci-libs/L-1": {"DEPEND": "sci-libs/K[-foo]", "EAPI": 2},
            "sci-libs/M-1": {"DEPEND": "sci-libs/K[foo=]", "IUSE": "+foo", "EAPI": 2},
        }

        installed = {}

        test_cases = (
            # Test bug 615824, where an automask USE change results in
            # a conflict which is not reported. In order to install L,
            # foo must be disabled for both K and M, but autounmask
            # disables foo for K and leaves it enabled for M.
            ResolverPlaygroundTestCase(
                ["sci-libs/L", "sci-libs/M"],
                options={"--backtrack": 0},
                success=False,
                mergelist=[
                    "sci-libs/L-1",
                    "sci-libs/M-1",
                    "sci-libs/K-1",
                ],
                ignore_mergelist_order=True,
                slot_collision_solutions=[
                    {"sci-libs/K-1": {"foo": False}, "sci-libs/M-1": {"foo": False}}
                ],
            ),
        )

        playground = ResolverPlayground(ebuilds=ebuilds, installed=installed)
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.debug = False
            playground.cleanup()



# Copyright 2015 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from __future__ import print_function
import sys

from portage.const import SUPPORTED_GENTOO_BINPKG_FORMATS
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)
from portage.output import colorize


class SonameDowngradeTestCase(TestCase):
    def testSingleSlot(self):

        ebuilds = {
            "dev-libs/icu-49": {},
            "dev-libs/icu-4.8": {},
            "dev-libs/libxml2-2.7.8": {
                "DEPEND": "dev-libs/icu",
                "RDEPEND": "dev-libs/icu",
            },
        }

        binpkgs = {
            "dev-libs/icu-49": {
                "PROVIDES": "x86_32: libicu.so.49",
            },
            "dev-libs/icu-4.8": {
                "PROVIDES": "x86_32: libicu.so.48",
            },
            "dev-libs/libxml2-2.7.8": {
                "DEPEND": "dev-libs/icu",
                "RDEPEND": "dev-libs/icu",
                "REQUIRES": "x86_32: libicu.so.48",
            },
        }
        installed = {
            "dev-libs/icu-49": {
                "PROVIDES": "x86_32: libicu.so.49",
            },
            "dev-libs/libxml2-2.7.8": {
                "DEPEND": "dev-libs/icu",
                "RDEPEND": "dev-libs/icu",
                "REQUIRES": "x86_32: libicu.so.49",
            },
        }

        user_config = {
            "package.mask": (">=dev-libs/icu-49",),
        }

        world = ["dev-libs/libxml2"]

        test_cases = (
            ResolverPlaygroundTestCase(
                ["dev-libs/icu"],
                options={
                    "--autounmask": "n",
                    "--ignore-soname-deps": "n",
                    "--oneshot": True,
                    "--usepkgonly": True,
                },
                success=True,
                mergelist=[
                    "[binary]dev-libs/icu-4.8",
                    "[binary]dev-libs/libxml2-2.7.8",
                ],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/icu"],
                options={
                    "--autounmask": "n",
                    "--ignore-soname-deps": "y",
                    "--oneshot": True,
                    "--usepkgonly": True,
                },
                success=True,
                mergelist=[
                    "[binary]dev-libs/icu-4.8",
                ],
            ),
            ResolverPlaygroundTestCase(
                ["@world"],
                options={
                    "--autounmask": "n",
                    "--deep": True,
                    "--ignore-soname-deps": "n",
                    "--update": True,
                    "--usepkgonly": True,
                },
                success=True,
                mergelist=[
                    "[binary]dev-libs/icu-4.8",
                    "[binary]dev-libs/libxml2-2.7.8",
                ],
            ),
            # In this case, soname dependencies are not respected,
            # because --usepkgonly is not enabled. This could be
            # handled differently, by respecting soname dependencies
            # as long as no unbuilt ebuilds get pulled into the graph.
            # However, that kind of conditional dependency accounting
            # would add a significant amount of complexity.
            ResolverPlaygroundTestCase(
                ["@world"],
                options={
                    "--deep": True,
                    "--ignore-soname-deps": "n",
                    "--update": True,
                    "--usepkg": True,
                },
                success=True,
                mergelist=[
                    "[binary]dev-libs/icu-4.8",
                ],
            ),
            ResolverPlaygroundTestCase(
                ["@world"],
                options={
                    "--deep": True,
                    "--update": True,
                },
                success=True,
                mergelist=[
                    "dev-libs/icu-4.8",
                ],
            ),
        )

        for binpkg_format in SUPPORTED_GENTOO_BINPKG_FORMATS:
            with self.subTest(binpkg_format=binpkg_format):
                print(colorize("HILITE", binpkg_format), end=" ... ")
                sys.stdout.flush()
                user_config["make.conf"] = ('BINPKG_FORMAT="%s"' % binpkg_format,)
                playground = ResolverPlayground(
                    binpkgs=binpkgs,
                    ebuilds=ebuilds,
                    installed=installed,
                    user_config=user_config,
                    world=world,
                    debug=False,
                )
                try:
                    for test_case in test_cases:
                        playground.run_TestCase(test_case)
                        self.assertEqual(
                            test_case.test_success, True, test_case.fail_msg
                        )
                finally:
                    # Disable debug so that cleanup works.
                    playground.debug = False
                    playground.cleanup()

    def testTwoSlots(self):

        ebuilds = {
            "dev-libs/glib-1.2.10": {"SLOT": "1"},
            "dev-libs/glib-2.30.2": {"SLOT": "2"},
            "dev-libs/dbus-glib-0.98": {
                "EAPI": "1",
                "DEPEND": "dev-libs/glib:2",
                "RDEPEND": "dev-libs/glib:2",
            },
        }
        binpkgs = {
            "dev-libs/glib-1.2.10": {
                "SLOT": "1",
                "PROVIDES": "x86_32: libglib-1.0.so.0",
            },
            "dev-libs/glib-2.30.2": {
                "PROVIDES": "x86_32: libglib-2.0.so.30",
                "SLOT": "2",
            },
            "dev-libs/glib-2.32.3": {
                "PROVIDES": "x86_32: libglib-2.0.so.32",
                "SLOT": "2",
            },
            "dev-libs/dbus-glib-0.98": {
                "EAPI": "1",
                "DEPEND": "dev-libs/glib:2",
                "RDEPEND": "dev-libs/glib:2",
                "REQUIRES": "x86_32: libglib-2.0.so.30",
            },
        }
        installed = {
            "dev-libs/glib-1.2.10": {
                "PROVIDES": "x86_32: libglib-1.0.so.0",
                "SLOT": "1",
            },
            "dev-libs/glib-2.32.3": {
                "PROVIDES": "x86_32: libglib-2.0.so.32",
                "SLOT": "2",
            },
            "dev-libs/dbus-glib-0.98": {
                "EAPI": "1",
                "DEPEND": "dev-libs/glib:2",
                "RDEPEND": "dev-libs/glib:2",
                "REQUIRES": "x86_32: libglib-2.0.so.32",
            },
        }

        user_config = {
            "package.mask": (">=dev-libs/glib-2.32",),
        }

        world = [
            "dev-libs/glib:1",
            "dev-libs/dbus-glib",
        ]

        test_cases = (
            ResolverPlaygroundTestCase(
                ["@world"],
                options={
                    "--autounmask": "n",
                    "--deep": True,
                    "--ignore-soname-deps": "n",
                    "--update": True,
                    "--usepkgonly": True,
                },
                success=True,
                mergelist=[
                    "[binary]dev-libs/glib-2.30.2",
                    "[binary]dev-libs/dbus-glib-0.98",
                ],
            ),
        )

        for binpkg_format in SUPPORTED_GENTOO_BINPKG_FORMATS:
            with self.subTest(binpkg_format=binpkg_format):
                print(colorize("HILITE", binpkg_format), end=" ... ")
                sys.stdout.flush()
                user_config["make.conf"] = ('BINPKG_FORMAT="%s"' % binpkg_format,)
                playground = ResolverPlayground(
                    ebuilds=ebuilds,
                    binpkgs=binpkgs,
                    installed=installed,
                    user_config=user_config,
                    world=world,
                    debug=False,
                )

                try:
                    for test_case in test_cases:
                        playground.run_TestCase(test_case)
                        self.assertEqual(
                            test_case.test_success, True, test_case.fail_msg
                        )
                finally:
                    # Disable debug so that cleanup works.
                    playground.debug = False
                    playground.cleanup()



# test_dobin.py -- Portage Unit Testing Functionality
# Copyright 2007-2010 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests.bin.setup_env import BinTestCase, dobin, xexists_in_D


class DoBin(BinTestCase):
    def testDoBin(self):
        self.init()
        try:
            dobin("does-not-exist", 1)
            xexists_in_D("does-not-exist")
            xexists_in_D("/bin/does-not-exist")
            xexists_in_D("/usr/bin/does-not-exist")
        finally:
            self.cleanup()



# Copyright 1999-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import datetime
import io
import logging
import random
import re
import signal
import socket
import sys
import tempfile
import time

from _emerge.UserQuery import UserQuery

import portage
from portage import _unicode_decode
from portage import os
from portage.const import VCS_DIRS, TIMESTAMP_FORMAT, RSYNC_PACKAGE_ATOM
from portage.output import create_color_func, yellow, blue, bold
from portage.sync.getaddrinfo_validate import getaddrinfo_validate
from portage.sync.syncbase import NewBase
from portage.util import writemsg, writemsg_level, writemsg_stdout
from portage.util.futures import asyncio

good = create_color_func("GOOD")
bad = create_color_func("BAD")
warn = create_color_func("WARN")

try:
    from gemato.exceptions import GematoException
    import gemato.openpgp
    import gemato.recursiveloader
except ImportError:
    gemato = None


SERVER_OUT_OF_DATE = -1
EXCEEDED_MAX_RETRIES = -2


class RsyncSync(NewBase):
    """Rsync sync module"""

    short_desc = "Perform sync operations on rsync based repositories"

    @staticmethod
    def name():
        return "RsyncSync"

    def __init__(self):
        NewBase.__init__(self, "rsync", RSYNC_PACKAGE_ATOM)

    def update(self):
        """Internal update function which performs the transfer"""
        opts = self.options.get("emerge_config").opts
        self.usersync_uid = self.options.get("usersync_uid", None)
        enter_invalid = "--ask-enter-invalid" in opts
        quiet = "--quiet" in opts
        out = portage.output.EOutput(quiet=quiet)
        syncuri = self.repo.sync_uri
        if self.repo.module_specific_options.get(
            "sync-rsync-vcs-ignore", "false"
        ).lower() in ("true", "yes"):
            vcs_dirs = ()
        else:
            vcs_dirs = frozenset(VCS_DIRS)
            vcs_dirs = vcs_dirs.intersection(os.listdir(self.repo.location))

        for vcs_dir in vcs_dirs:
            writemsg_level(
                (
                    "!!! %s appears to be under revision "
                    + "control (contains %s).\n!!! Aborting rsync sync "
                    '(override with "sync-rsync-vcs-ignore = true" in repos.conf).\n'
                )
                % (self.repo.location, vcs_dir),
                level=logging.ERROR,
                noiselevel=-1,
            )
            return (1, False)
        self.timeout = 180

        rsync_opts = []
        if self.settings["PORTAGE_RSYNC_OPTS"] == "":
            rsync_opts = self._set_rsync_defaults()
        else:
            rsync_opts = self._validate_rsync_opts(rsync_opts, syncuri)
        self.rsync_opts = self._rsync_opts_extend(opts, rsync_opts)

        self.extra_rsync_opts = list()
        if self.repo.module_specific_options.get("sync-rsync-extra-opts"):
            self.extra_rsync_opts.extend(
                portage.util.shlex_split(
                    self.repo.module_specific_options["sync-rsync-extra-opts"]
                )
            )

        exitcode = 0
        verify_failure = False

        # Process GLEP74 verification options.
        # Default verification to 'no'; it's enabled for ::gentoo
        # via default repos.conf though.
        self.verify_metamanifest = self.repo.module_specific_options.get(
            "sync-rsync-verify-metamanifest", "no"
        ).lower() in ("yes", "true")
        # Support overriding job count.
        self.verify_jobs = self.repo.module_specific_options.get(
            "sync-rsync-verify-jobs", None
        )
        if self.verify_jobs is not None:
            try:
                self.verify_jobs = int(self.verify_jobs)
                if self.verify_jobs < 0:
                    raise ValueError(self.verify_jobs)
            except ValueError:
                writemsg_level(
                    "!!! sync-rsync-verify-jobs not a positive integer: %s\n"
                    % (self.verify_jobs,),
                    level=logging.WARNING,
                    noiselevel=-1,
                )
                self.verify_jobs = None
            else:
                if self.verify_jobs == 0:
                    # Use the apparent number of processors if gemato
                    # supports it.
                    self.verify_jobs = None
        # Support overriding max age.
        self.max_age = self.repo.module_specific_options.get(
            "sync-rsync-verify-max-age", ""
        )
        if self.max_age:
            try:
                self.max_age = int(self.max_age)
                if self.max_age < 0:
                    raise ValueError(self.max_age)
            except ValueError:
                writemsg_level(
                    "!!! sync-rsync-max-age must be a non-negative integer: %s\n"
                    % (self.max_age,),
                    level=logging.WARNING,
                    noiselevel=-1,
                )
                self.max_age = 0
        else:
            self.max_age = 0

        openpgp_env = None
        if self.verify_metamanifest and gemato is not None:
            # Use isolated environment if key is specified,
            # system environment otherwise
            openpgp_env = self._get_openpgp_env(self.repo.sync_openpgp_key_path)

        try:
            # Load and update the keyring early. If it fails, then verification
            # will not be performed and the user will have to fix it and try again,
            # so we may as well bail out before actual rsync happens.
            if openpgp_env is not None and self.repo.sync_openpgp_key_path is not None:
                try:
                    out.einfo("Using keys from %s" % (self.repo.sync_openpgp_key_path,))
                    with io.open(self.repo.sync_openpgp_key_path, "rb") as f:
                        openpgp_env.import_key(f)
                    self._refresh_keys(openpgp_env)
                except (GematoException, asyncio.TimeoutError) as e:
                    writemsg_level(
                        "!!! Manifest verification impossible due to keyring problem:\n%s\n"
                        % (e,),
                        level=logging.ERROR,
                        noiselevel=-1,
                    )
                    return (1, False)

            # Real local timestamp file.
            self.servertimestampfile = os.path.join(
                self.repo.location, "metadata", "timestamp.chk"
            )

            content = portage.util.grabfile(self.servertimestampfile)
            timestamp = 0
            if content:
                try:
                    timestamp = time.mktime(time.strptime(content[0], TIMESTAMP_FORMAT))
                except (OverflowError, ValueError):
                    pass
            del content

            try:
                self.rsync_initial_timeout = int(
                    self.settings.get("PORTAGE_RSYNC_INITIAL_TIMEOUT", "15")
                )
            except ValueError:
                self.rsync_initial_timeout = 15

            try:
                maxretries = int(self.settings["PORTAGE_RSYNC_RETRIES"])
            except SystemExit as e:
                raise  # Needed else can't exit
            except:
                maxretries = -1  # default number of retries

            if syncuri.startswith("file://"):
                self.proto = "file"
                dosyncuri = syncuri[7:]
                unchanged, is_synced, exitcode, updatecache_flg = self._do_rsync(
                    dosyncuri, timestamp, opts
                )
                self._process_exitcode(exitcode, dosyncuri, out, 1)
                if exitcode == 0:
                    if unchanged:
                        self.repo_storage.abort_update()
                    else:
                        self.repo_storage.commit_update()
                        self.repo_storage.garbage_collection()
                return (exitcode, updatecache_flg)

            retries = 0
            try:
                self.proto, user_name, hostname, port = re.split(
                    r"(rsync|ssh)://([^:/]+@)?(\[[:\da-fA-F]*\]|[^:/]*)(:[0-9]+)?",
                    syncuri,
                    maxsplit=4,
                )[1:5]
            except ValueError:
                writemsg_level(
                    "!!! sync-uri is invalid: %s\n" % syncuri,
                    noiselevel=-1,
                    level=logging.ERROR,
                )
                return (1, False)

            self.ssh_opts = self.settings.get("PORTAGE_SSH_OPTS")

            if port is None:
                port = ""
            if user_name is None:
                user_name = ""
            if re.match(r"^\[[:\da-fA-F]*\]$", hostname) is None:
                getaddrinfo_host = hostname
            else:
                # getaddrinfo needs the brackets stripped
                getaddrinfo_host = hostname[1:-1]
            updatecache_flg = False
            all_rsync_opts = set(self.rsync_opts)
            all_rsync_opts.update(self.extra_rsync_opts)

            family = socket.AF_UNSPEC
            if "-4" in all_rsync_opts or "--ipv4" in all_rsync_opts:
                family = socket.AF_INET
            elif socket.has_ipv6 and (
                "-6" in all_rsync_opts or "--ipv6" in all_rsync_opts
            ):
                family = socket.AF_INET6

            addrinfos = None
            uris = []

            if "RSYNC_PROXY" not in self.spawn_kwargs["env"]:
                try:
                    addrinfos = getaddrinfo_validate(
                        socket.getaddrinfo(
                            getaddrinfo_host, None, family, socket.SOCK_STREAM
                        )
                    )
                except socket.error as e:
                    writemsg_level(
                        "!!! getaddrinfo failed for '%s': %s\n"
                        % (_unicode_decode(hostname), str(e)),
                        noiselevel=-1,
                        level=logging.ERROR,
                    )

            if addrinfos:

                AF_INET = socket.AF_INET
                AF_INET6 = None
                if socket.has_ipv6:
                    AF_INET6 = socket.AF_INET6

                ips_v4 = []
                ips_v6 = []

                for addrinfo in addrinfos:
                    if addrinfo[0] == AF_INET:
                        ips_v4.append("%s" % addrinfo[4][0])
                    elif AF_INET6 is not None and addrinfo[0] == AF_INET6:
                        # IPv6 addresses need to be enclosed in square brackets
                        ips_v6.append("[%s]" % addrinfo[4][0])

                random.shuffle(ips_v4)
                random.shuffle(ips_v6)

                # Give priority to the address family that
                # getaddrinfo() returned first.
                if AF_INET6 is not None and addrinfos and addrinfos[0][0] == AF_INET6:
                    ips = ips_v6 + ips_v4
                else:
                    ips = ips_v4 + ips_v6

                for ip in ips:
                    uris.append(
                        syncuri.replace(
                            "//" + user_name + hostname + port + "/",
                            "//" + user_name + ip + port + "/",
                            1,
                        )
                    )

            if not uris:
                # With some configurations we need to use the plain hostname
                # rather than try to resolve the ip addresses (bug #340817).
                uris.append(syncuri)
            elif len(uris) == 1:
                # Use the original hostname if it resolves to a single IP,
                # since DNS lookup must occur in the rsync process for
                # compatibility with things like proxychains that allocate
                # a surrogate IP which is only valid within the current
                # process.
                uris = [syncuri]

            # reverse, for use with pop()
            uris.reverse()
            uris_orig = uris[:]

            effective_maxretries = maxretries
            if effective_maxretries < 0:
                effective_maxretries = len(uris) - 1

            local_state_unchanged = True
            while 1:
                if uris:
                    dosyncuri = uris.pop()
                elif maxretries < 0 or retries > maxretries:
                    writemsg(
                        "!!! Exhausted addresses for %s\n" % _unicode_decode(hostname),
                        noiselevel=-1,
                    )
                    return (1, False)
                else:
                    uris.extend(uris_orig)
                    dosyncuri = uris.pop()

                if retries == 0:
                    if "--ask" in opts:
                        uq = UserQuery(opts)
                        if (
                            uq.query(
                                "Do you want to sync your ebuild repository "
                                + "with the mirror at\n"
                                + blue(dosyncuri)
                                + bold("?"),
                                enter_invalid,
                            )
                            == "No"
                        ):
                            print()
                            print("Quitting.")
                            print()
                            sys.exit(128 + signal.SIGINT)
                    self.logger(
                        self.xterm_titles, ">>> Starting rsync with " + dosyncuri
                    )
                    if "--quiet" not in opts:
                        print(">>> Starting rsync with " + dosyncuri + "...")
                else:
                    self.logger(
                        self.xterm_titles,
                        ">>> Starting retry %d of %d with %s"
                        % (retries, effective_maxretries, dosyncuri),
                    )
                    writemsg_stdout(
                        "\n\n>>> Starting retry %d of %d with %s\n"
                        % (retries, effective_maxretries, dosyncuri),
                        noiselevel=-1,
                    )

                if dosyncuri.startswith("ssh://"):
                    dosyncuri = dosyncuri[6:].replace("/", ":/", 1)

                unchanged, is_synced, exitcode, updatecache_flg = self._do_rsync(
                    dosyncuri, timestamp, opts
                )
                if not unchanged:
                    local_state_unchanged = False
                if is_synced:
                    break

                retries = retries + 1

                if maxretries < 0 or retries <= maxretries:
                    print(">>> Retrying...")
                else:
                    # over retries
                    # exit loop
                    exitcode = EXCEEDED_MAX_RETRIES
                    break

            self._process_exitcode(exitcode, dosyncuri, out, maxretries)

            if local_state_unchanged:
                # The quarantine download_dir is not intended to exist
                # in this case, so refer gemato to the normal repository
                # location.
                download_dir = self.repo.location
            else:
                download_dir = self.download_dir

            # if synced successfully, verify now
            if exitcode == 0 and self.verify_metamanifest:
                if gemato is None:
                    writemsg_level(
                        "!!! Unable to verify: gemato-14.5+ is required\n",
                        level=logging.ERROR,
                        noiselevel=-1,
                    )
                    exitcode = 127
                else:
                    try:
                        # we always verify the Manifest signature, in case
                        # we had to deal with key revocation case
                        m = gemato.recursiveloader.ManifestRecursiveLoader(
                            os.path.join(download_dir, "Manifest"),
                            verify_openpgp=True,
                            openpgp_env=openpgp_env,
                            max_jobs=self.verify_jobs,
                        )
                        if not m.openpgp_signed:
                            raise RuntimeError(
                                "OpenPGP signature not found on Manifest"
                            )

                        ts = m.find_timestamp()
                        if ts is None:
                            raise RuntimeError("Timestamp not found in Manifest")
                        if (
                            self.max_age != 0
                            and (datetime.datetime.utcnow() - ts.ts).days > self.max_age
                        ):
                            out.quiet = False
                            out.ewarn(
                                "Manifest is over %d days old, this is suspicious!"
                                % (self.max_age,)
                            )
                            out.ewarn(
                                "You may want to try using another mirror and/or reporting this one:"
                            )
                            out.ewarn("  %s" % (dosyncuri,))
                            out.ewarn("")
                            out.quiet = quiet

                        out.einfo("Manifest timestamp: %s UTC" % (ts.ts,))
                        out.einfo("Valid OpenPGP signature found:")
                        out.einfo(
                            "- primary key: %s"
                            % (m.openpgp_signature.primary_key_fingerprint)
                        )
                        out.einfo("- subkey: %s" % (m.openpgp_signature.fingerprint))
                        out.einfo(
                            "- timestamp: %s UTC" % (m.openpgp_signature.timestamp)
                        )

                        # if nothing has changed, skip the actual Manifest
                        # verification
                        if not local_state_unchanged:
                            out.ebegin("Verifying %s" % (download_dir,))
                            m.assert_directory_verifies()
                            out.eend(0)
                    except GematoException as e:
                        writemsg_level(
                            "!!! Manifest verification failed:\n%s\n" % (e,),
                            level=logging.ERROR,
                            noiselevel=-1,
                        )
                        exitcode = 1
                        verify_failure = True

            if exitcode == 0 and not local_state_unchanged:
                self.repo_storage.commit_update()
                self.repo_storage.garbage_collection()

            return (exitcode, updatecache_flg)
        finally:
            # Don't delete the update if verification failed, in case
            # the cause needs to be investigated.
            if not verify_failure:
                self.repo_storage.abort_update()
            if openpgp_env is not None:
                openpgp_env.close()

    def _process_exitcode(self, exitcode, syncuri, out, maxretries):
        if exitcode == 0:
            pass
        elif exitcode == SERVER_OUT_OF_DATE:
            exitcode = 1
        elif exitcode == EXCEEDED_MAX_RETRIES:
            sys.stderr.write(">>> Exceeded PORTAGE_RSYNC_RETRIES: %s\n" % maxretries)
            exitcode = 1
        elif exitcode > 0:
            msg = []
            if exitcode == 1:
                msg.append(
                    "Rsync has reported that there is a syntax error. Please ensure"
                )
                msg.append(
                    "that sync-uri attribute for repository '%s' is proper."
                    % self.repo.name
                )
                msg.append("sync-uri: '%s'" % self.repo.sync_uri)
            elif exitcode == 11:
                msg.append("Rsync has reported that there is a File IO error. Normally")
                msg.append(
                    "this means your disk is full, but can be caused by corruption"
                )
                msg.append(
                    "on the filesystem that contains repository '%s'. Please investigate"
                    % self.repo.name
                )
                msg.append("and try again after the problem has been fixed.")
                msg.append("Location of repository: '%s'" % self.repo.location)
            elif exitcode == 20:
                msg.append("Rsync was killed before it finished.")
            else:
                msg.append(
                    "Rsync has not successfully finished. It is recommended that you keep"
                )
                msg.append(
                    "trying or that you use the 'emerge-webrsync' option if you are unable"
                )
                msg.append(
                    "to use rsync due to firewall or other restrictions. This should be a"
                )
                msg.append(
                    "temporary problem unless complications exist with your network"
                )
                msg.append("(and possibly your system's filesystem) configuration.")
            for line in msg:
                out.eerror(line)

    def new(self, **kwargs):
        if kwargs:
            self._kwargs(kwargs)
        try:
            if not os.path.exists(self.repo.location):
                os.makedirs(self.repo.location)
                self.logger(
                    self.self.xterm_titles,
                    "Created New Directory %s " % self.repo.location,
                )
        except IOError:
            return (1, False)
        return self.update()

    def retrieve_head(self, **kwargs):
        """Get information about the head commit"""
        if kwargs:
            self._kwargs(kwargs)
        last_sync = portage.grabfile(
            os.path.join(self.repo.location, "metadata", "timestamp.commit")
        )
        ret = (1, False)
        if last_sync:
            try:
                ret = (os.EX_OK, last_sync[0].split()[0])
            except IndexError:
                pass
        return ret

    def _set_rsync_defaults(self):
        portage.writemsg(
            "PORTAGE_RSYNC_OPTS empty or unset, using hardcoded defaults\n"
        )
        rsync_opts = [
            "--recursive",  # Recurse directories
            "--links",  # Consider symlinks
            "--safe-links",  # Ignore links outside of tree
            "--perms",  # Preserve permissions
            "--times",  # Preserive mod times
            "--omit-dir-times",
            "--compress",  # Compress the data transmitted
            "--force",  # Force deletion on non-empty dirs
            "--whole-file",  # Don't do block transfers, only entire files
            "--delete",  # Delete files that aren't in the master tree
            "--stats",  # Show final statistics about what was transfered
            "--human-readable",
            "--timeout=" + str(self.timeout),  # IO timeout if not done in X seconds
            "--exclude=/distfiles",  # Exclude distfiles from consideration
            "--exclude=/local",  # Exclude local     from consideration
            "--exclude=/packages",  # Exclude packages  from consideration
        ]
        return rsync_opts

    def _validate_rsync_opts(self, rsync_opts, syncuri):
        # The below validation is not needed when using the above hardcoded
        # defaults.

        portage.writemsg("Using PORTAGE_RSYNC_OPTS instead of hardcoded defaults\n", 1)
        rsync_opts.extend(
            portage.util.shlex_split(self.settings.get("PORTAGE_RSYNC_OPTS", ""))
        )
        for opt in ("--recursive", "--times"):
            if opt not in rsync_opts:
                portage.writemsg(
                    yellow("WARNING:")
                    + " adding required option "
                    + "%s not included in PORTAGE_RSYNC_OPTS\n" % opt
                )
                rsync_opts.append(opt)

        for exclude in ("distfiles", "local", "packages"):
            opt = "--exclude=/%s" % exclude
            if opt not in rsync_opts:
                portage.writemsg(
                    yellow("WARNING:")
                    + " adding required option %s not included in " % opt
                    + "PORTAGE_RSYNC_OPTS (can be overridden with --exclude='!')\n"
                )
                rsync_opts.append(opt)

        if syncuri.rstrip("/").endswith(".gentoo.org/gentoo-portage"):

            def rsync_opt_startswith(opt_prefix):
                for x in rsync_opts:
                    if x.startswith(opt_prefix):
                        return (1, False)
                return (0, False)

            if not rsync_opt_startswith("--timeout="):
                rsync_opts.append("--timeout=%d" % self.timeout)

            for opt in ("--compress", "--whole-file"):
                if opt not in rsync_opts:
                    portage.writemsg(
                        yellow("WARNING:")
                        + " adding required option "
                        + "%s not included in PORTAGE_RSYNC_OPTS\n" % opt
                    )
                    rsync_opts.append(opt)
        return rsync_opts

    @staticmethod
    def _rsync_opts_extend(opts, rsync_opts):
        if "--quiet" in opts:
            rsync_opts.append("--quiet")  # Shut up a lot
        else:
            rsync_opts.append("--verbose")  # Print filelist

        if "--verbose" in opts:
            rsync_opts.append("--progress")  # Progress meter for each file

        if "--debug" in opts:
            rsync_opts.append("--checksum")  # Force checksum on all files
        return rsync_opts

    def _do_rsync(self, syncuri, timestamp, opts):
        updatecache_flg = False
        is_synced = False
        if timestamp != 0 and "--quiet" not in opts:
            print(">>> Checking server timestamp ...")

        rsynccommand = [self.bin_command] + self.rsync_opts + self.extra_rsync_opts

        if self.proto == "ssh" and self.ssh_opts:
            rsynccommand.append("--rsh=ssh " + self.ssh_opts)

        if "--debug" in opts:
            print(rsynccommand)

        local_state_unchanged = False
        exitcode = os.EX_OK
        servertimestamp = 0
        # Even if there's no timestamp available locally, fetch the
        # timestamp anyway as an initial probe to verify that the server is
        # responsive.  This protects us from hanging indefinitely on a
        # connection attempt to an unresponsive server which rsync's
        # --timeout option does not prevent.

        # if True:
        # Temporary file for remote server timestamp comparison.
        # NOTE: If FEATURES=usersync is enabled then the tempfile
        # needs to be in a directory that's readable by the usersync
        # user. We assume that ${PORTAGE_TMPDIR}/portage will satisfy this
        # requirement, since that's not necessarily true for the
        # default directory used by the tempfile module.
        if self.usersync_uid is not None:
            tmpdir = os.path.join(self.settings["PORTAGE_TMPDIR"], "portage")
            ensure_dirs_kwargs = {}
            if portage.secpass >= 1:
                ensure_dirs_kwargs["gid"] = portage.portage_gid
                ensure_dirs_kwargs["mode"] = 0o70
                ensure_dirs_kwargs["mask"] = 0
            portage.util.ensure_dirs(tmpdir, **ensure_dirs_kwargs)
        else:
            # use default dir from tempfile module
            tmpdir = None
        fd, tmpservertimestampfile = tempfile.mkstemp(dir=tmpdir)
        os.close(fd)
        if self.usersync_uid is not None:
            portage.util.apply_permissions(
                tmpservertimestampfile, uid=self.usersync_uid
            )
        command = rsynccommand[:]
        command.append("--inplace")
        command.append(syncuri.rstrip("/") + "/metadata/timestamp.chk")
        command.append(tmpservertimestampfile)
        content = None
        pids = []
        try:
            # Timeout here in case the server is unresponsive.  The
            # --timeout rsync option doesn't apply to the initial
            # connection attempt.
            try:
                if self.rsync_initial_timeout:
                    portage.exception.AlarmSignal.register(self.rsync_initial_timeout)

                pids.extend(
                    portage.process.spawn(command, returnpid=True, **self.spawn_kwargs)
                )
                exitcode = os.waitpid(pids[0], 0)[1]
                if self.usersync_uid is not None:
                    portage.util.apply_permissions(
                        tmpservertimestampfile, uid=os.getuid()
                    )
                content = portage.grabfile(tmpservertimestampfile)
            finally:
                if self.rsync_initial_timeout:
                    portage.exception.AlarmSignal.unregister()
                try:
                    os.unlink(tmpservertimestampfile)
                except OSError:
                    pass
        except portage.exception.AlarmSignal:
            # timed out
            print("timed out")
            # With waitpid and WNOHANG, only check the
            # first element of the tuple since the second
            # element may vary (bug #337465).
            if pids and os.waitpid(pids[0], os.WNOHANG)[0] == 0:
                os.kill(pids[0], signal.SIGTERM)
                os.waitpid(pids[0], 0)
            # This is the same code rsync uses for timeout.
            exitcode = 30
        else:
            if exitcode != os.EX_OK:
                if exitcode & 0xFF:
                    exitcode = (exitcode & 0xFF) << 8
                else:
                    exitcode = exitcode >> 8

        if content:
            try:
                servertimestamp = time.mktime(
                    time.strptime(content[0], TIMESTAMP_FORMAT)
                )
            except (OverflowError, ValueError):
                pass
        del command, pids, content

        if exitcode == os.EX_OK:
            if (servertimestamp != 0) and (servertimestamp == timestamp):
                local_state_unchanged = True
                is_synced = True
                self.logger(
                    self.xterm_titles, ">>> Cancelling sync -- Already current."
                )
                print()
                print(">>>")
                print(
                    ">>> Timestamps on the server and in the local repository are the same."
                )
                print(
                    ">>> Cancelling all further sync action. You are already up to date."
                )
                print(">>>")
                print(
                    ">>> In order to force sync, remove '%s'."
                    % self.servertimestampfile
                )
                print(">>>")
                print()
            elif (servertimestamp != 0) and (servertimestamp < timestamp):
                self.logger(self.xterm_titles, ">>> Server out of date: %s" % syncuri)
                print()
                print(">>>")
                print(">>> SERVER OUT OF DATE: %s" % syncuri)
                print(">>>")
                print(
                    ">>> In order to force sync, remove '%s'."
                    % self.servertimestampfile
                )
                print(">>>")
                print()
                exitcode = SERVER_OUT_OF_DATE
            elif (servertimestamp == 0) or (servertimestamp > timestamp):
                # actual sync
                command = rsynccommand[:]

                submodule_paths = self._get_submodule_paths()
                if submodule_paths:
                    # The only way to select multiple directories to
                    # sync, without calling rsync multiple times, is
                    # to use --relative.
                    command.append("--relative")
                    for path in submodule_paths:
                        # /./ is special syntax supported with the
                        # rsync --relative option.
                        command.append(syncuri + "/./" + path)
                else:
                    command.append(syncuri + "/")

                command.append(self.download_dir)

                exitcode = None
                try:
                    exitcode = portage.process.spawn(command, **self.spawn_kwargs)
                finally:
                    if exitcode is None:
                        # interrupted
                        exitcode = 128 + signal.SIGINT

                    #   0	Success
                    #   1	Syntax or usage error
                    #   2	Protocol incompatibility
                    #   5	Error starting client-server protocol
                    #  35	Timeout waiting for daemon connection
                    if exitcode not in (0, 1, 2, 5, 35):
                        # If the exit code is not among those listed above,
                        # then we may have a partial/inconsistent sync
                        # state, so our previously read timestamp as well
                        # as the corresponding file can no longer be
                        # trusted.
                        timestamp = 0
                        try:
                            os.unlink(self.servertimestampfile)
                        except OSError:
                            pass
                    else:
                        updatecache_flg = True

                if exitcode in [0, 1, 3, 4, 11, 14, 20, 21]:
                    is_synced = True
        elif exitcode in [1, 3, 4, 11, 14, 20, 21]:
            is_synced = True
        else:
            # Code 2 indicates protocol incompatibility, which is expected
            # for servers with protocol < 29 that don't support
            # --prune-empty-directories.  Retry for a server that supports
            # at least rsync protocol version 29 (>=rsync-2.6.4).
            pass

        return local_state_unchanged, is_synced, exitcode, updatecache_flg



# Copyright 2014 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class SlotConflictForceRebuildTestCase(TestCase):
    def testSlotConflictForceRebuild(self):

        ebuilds = {
            "app-misc/A-1": {"EAPI": "5", "SLOT": "0/1"},
            "app-misc/A-2": {"EAPI": "5", "SLOT": "0/2"},
            "app-misc/B-0": {"EAPI": "5", "RDEPEND": "app-misc/A:="},
            "app-misc/C-0": {"EAPI": "5", "RDEPEND": "app-misc/A"},
        }

        installed = {
            "app-misc/A-1": {"EAPI": "5", "SLOT": "0/1"},
            "app-misc/B-0": {"EAPI": "5", "RDEPEND": "app-misc/A:0/1="},
            "app-misc/C-0": {"EAPI": "5", "RDEPEND": "app-misc/A:0/1="},
        }

        world = ["app-misc/B", "app-misc/C"]

        test_cases = (
            # Test bug #521990, where forced_rebuilds omits ebuilds that
            # had have had their slot operator atoms removed from the
            # ebuilds, even though the corresponding installed
            # instances had really forced rebuilds due to being built
            # with slot-operators in their deps.
            ResolverPlaygroundTestCase(
                ["app-misc/A"],
                options={},
                success=True,
                ambiguous_merge_order=True,
                mergelist=["app-misc/A-2", ("app-misc/B-0", "app-misc/C-0")],
                forced_rebuilds={"app-misc/A-2": ["app-misc/B-0", "app-misc/C-0"]},
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world, debug=False
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



# Copyright 2003-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import codecs
import operator
import portage
import re
import sys
import xml.dom.minidom

from functools import reduce

import io
from io import StringIO

from portage import _encodings, _unicode_decode, _unicode_encode
from portage import os
from portage.const import PRIVATE_PATH
from portage.dep import _slot_separator
from portage.localization import _
from portage.util import grabfile
from portage.versions import pkgsplit, vercmp

from urllib.request import urlopen as urllib_request_urlopen

# Note: the space for rgt and rlt is important !!
# FIXME: use slot deps instead, requires GLSA format versioning
opMapping = {
    "le": "<=",
    "lt": "<",
    "eq": "=",
    "gt": ">",
    "ge": ">=",
    "rge": ">=~",
    "rle": "<=~",
    "rgt": " >~",
    "rlt": " <~",
}
NEWLINE_ESCAPE = "!;\\n"  # some random string to mark newlines that should be preserved
SPACE_ESCAPE = "!;_"  # some random string to mark spaces that should be preserved


def get_applied_glsas(settings):
    """
    Return a list of applied or injected GLSA IDs

    @type	settings: portage.config
    @param	settings: portage config instance
    @rtype:		list
    @return:	list of glsa IDs
    """
    return grabfile(os.path.join(settings["EROOT"], PRIVATE_PATH, "glsa_injected"))


# TODO: use the textwrap module instead
def wrap(text, width, caption=""):
    """
    Wraps the given text at column I{width}, optionally indenting
    it so that no text is under I{caption}. It's possible to encode
    hard linebreaks in I{text} with L{NEWLINE_ESCAPE}.

    @type	text: String
    @param	text: the text to be wrapped
    @type	width: Integer
    @param	width: the column at which the text should be wrapped
    @type	caption: String
    @param	caption: this string is inserted at the beginning of the
                                     return value and the paragraph is indented up to
                                     C{len(caption)}.
    @rtype:		String
    @return:	the wrapped and indented paragraph
    """
    rValue = ""
    line = caption
    text = text.replace(2 * NEWLINE_ESCAPE, NEWLINE_ESCAPE + " " + NEWLINE_ESCAPE)
    words = text.split()
    indentLevel = len(caption) + 1

    for word in words:
        if line and line[-1] == "\n":
            rValue = f"{rValue}{line}"
            line = " " * indentLevel
        if len(line) + len(word.replace(NEWLINE_ESCAPE, "")) + 1 > width:
            rValue = f"{rValue}{line}\n"
            escaped_word = word.replace(NEWLINE_ESCAPE, "\n")
            line = f"{' ' * indentLevel}{escaped_word}"
        elif word.find(NEWLINE_ESCAPE) >= 0:
            escaped_word = word.replace(NEWLINE_ESCAPE, "\n")
            whitespace = ""
            if len(line.strip()) > 0:
                whitespace = " "
            rValue = f"{rValue}{line}{whitespace}{escaped_word}"
            line = " " * indentLevel
        else:
            whitespace = ""
            if len(line.strip()) > 0:
                whitespace = " "
            line = f"{line}{whitespace}{word}"
    if len(line) > 0:
        escaped_line = line.replace(NEWLINE_ESCAPE, "\n")
        rValue = f"{rValue}{escaped_line}"
    rValue = rValue.replace(SPACE_ESCAPE, " ")
    return rValue


def get_glsa_list(myconfig):
    """
    Returns a list of all available GLSAs in the given repository
    by comparing the filelist there with the pattern described in
    the config.

    @type	myconfig: portage.config
    @param	myconfig: Portage settings instance

    @rtype:		List of Strings
    @return:	a list of GLSA IDs in this repository
    """

    repository = os.path.join(myconfig["PORTDIR"], "metadata", "glsa")
    if "GLSA_DIR" in myconfig:
        repository = myconfig["GLSA_DIR"]

    if not os.access(repository, os.R_OK):
        return []
    dirlist = os.listdir(repository)
    prefix = "glsa-"
    prefix_size = len(prefix)
    suffix = ".xml"
    suffix_size = len(suffix)

    def check(value):
        try:
            if value[:prefix_size] == prefix and value[-suffix_size:] == suffix:
                return value[prefix_size:-suffix_size]
        except IndexError:
            return None
        return None

    checked_dirlist = (check(f) for f in dirlist)
    rValue = [f for f in checked_dirlist if f]
    return rValue


def getListElements(listnode):
    """
    Get all <li> elements for a given <ol> or <ul> node.

    @type	listnode: xml.dom.Node
    @param	listnode: <ul> or <ol> list to get the elements for
    @rtype:		List of Strings
    @return:	a list that contains the value of the <li> elements
    """
    if not listnode.nodeName in ("ul", "ol"):
        raise GlsaFormatException("Invalid function call: listnode is not <ul> or <ol>")
    rValue = [
        getText(li, format="strip")
        for li in listnode.childNodes
        if li.nodeType == xml.dom.Node.ELEMENT_NODE
    ]
    return rValue


def getText(node, format, textfd=None):  # pylint: disable=redefined-builtin
    """
    This is the main parser function. It takes a node and traverses
    recursive over the subnodes, getting the text of each (and the
    I{link} attribute for <uri> and <mail>). Depending on the I{format}
    parameter the text might be formatted by adding/removing newlines,
    tabs and spaces. This function is only useful for the GLSA DTD,
    it's not applicable for other DTDs.

    @type	node: xml.dom.Node
    @param	node: the root node to start with the parsing
    @type	format: String
    @param	format: this should be either I{strip}, I{keep} or I{xml}
                                    I{keep} just gets the text and does no formatting.
                                    I{strip} replaces newlines and tabs with spaces and
                                    replaces multiple spaces with one space.
                                    I{xml} does some more formatting, depending on the
                                    type of the encountered nodes.
    @type	textfd: writable file-like object
    @param	textfd: the file-like object to write the output to
    @rtype:		String
    @return:	the (formatted) content of the node and its subnodes
                    except if textfd was not none
    """
    if not textfd:
        textfd = StringIO()
        returnNone = False
    else:
        returnNone = True
    if format in ("strip", "keep"):
        if node.nodeName in ("uri", "mail"):
            textfd.write(f"{node.childNodes[0].data}:{node.getAttribute('link')}")
        else:
            for subnode in node.childNodes:
                if subnode.nodeName == "#text":
                    textfd.write(subnode.data)
                else:
                    getText(subnode, format, textfd)
    else:  # format = "xml"
        for subnode in node.childNodes:
            if subnode.nodeName == "p":
                for p_subnode in subnode.childNodes:
                    if p_subnode.nodeName == "#text":
                        textfd.write(p_subnode.data.strip())
                    elif p_subnode.nodeName in ("uri", "mail"):
                        textfd.write(p_subnode.childNodes[0].data)
                        textfd.write(" ( " + p_subnode.getAttribute("link") + " )")
                textfd.write(NEWLINE_ESCAPE)
            elif subnode.nodeName == "ul":
                for li in getListElements(subnode):
                    textfd.write(f"-{SPACE_ESCAPE}{li}{NEWLINE_ESCAPE} ")
            elif subnode.nodeName == "ol":
                i = 0
                for li in getListElements(subnode):
                    i = i + 1
                    textfd.write(f"{i}.{SPACE_ESCAPE}{li}{NEWLINE_ESCAPE} ")
            elif subnode.nodeName == "code":
                textfd.write(
                    getText(subnode, format="keep")
                    .lstrip()
                    .replace("\n", NEWLINE_ESCAPE)
                )
                textfd.write(NEWLINE_ESCAPE)
            elif subnode.nodeName == "#text":
                textfd.write(subnode.data)
            else:
                raise GlsaFormatException(_("Invalid Tag found: "), subnode.nodeName)
    if returnNone:
        return None
    rValue = textfd.getvalue()
    if format == "strip":
        rValue = rValue.strip(" \n\t")
        rValue = re.sub(r"[\s]{2,}", " ", rValue)
    return rValue


def getMultiTagsText(rootnode, tagname, format):  # pylint: disable=redefined-builtin
    """
    Returns a list with the text of all subnodes of type I{tagname}
    under I{rootnode} (which itself is not parsed) using the given I{format}.

    @type	rootnode: xml.dom.Node
    @param	rootnode: the node to search for I{tagname}
    @type	tagname: String
    @param	tagname: the name of the tags to search for
    @type	format: String
    @param	format: see L{getText}
    @rtype:		List of Strings
    @return:	a list containing the text of all I{tagname} childnodes
    """
    rValue = [getText(e, format) for e in rootnode.getElementsByTagName(tagname)]
    return rValue


def makeAtom(pkgname, versionNode):
    """
    creates from the given package name and information in the
    I{versionNode} a (syntactical) valid portage atom.

    @type	pkgname: String
    @param	pkgname: the name of the package for this atom
    @type	versionNode: xml.dom.Node
    @param	versionNode: a <vulnerable> or <unaffected> Node that
                                             contains the version information for this atom
    @rtype:		String
    @return:	the portage atom
    """
    op = opMapping[versionNode.getAttribute("range")]
    version = getText(versionNode, format="strip")
    rValue = f"{op}{pkgname}-{version}"
    try:
        slot = versionNode.getAttribute("slot").strip()
    except KeyError:
        pass
    else:
        if slot and slot != "*":
            rValue = f"{rValue}{_slot_separator}{slot}"
    return str(rValue)


def makeVersion(versionNode):
    """
    creates from the information in the I{versionNode} a
    version string (format <op><version>).

    @type	versionNode: xml.dom.Node
    @param	versionNode: a <vulnerable> or <unaffected> Node that
                                             contains the version information for this atom
    @rtype:		String
    @return:	the version string
    """
    op = opMapping[versionNode.getAttribute("range")]
    version = getText(versionNode, format="strip")
    rValue = f"{op}{version}"
    try:
        slot = versionNode.getAttribute("slot").strip()
    except KeyError:
        pass
    else:
        if slot and slot != "*":
            rValue = f"{rValue}{_slot_separator}{slot}"
    return rValue


def match(atom, dbapi, match_type="default"):
    """
    wrapper that calls revisionMatch() or portage.dbapi.dbapi.match() depending on
    the given atom.

    @type	atom: string
    @param	atom: a <~ or >~ atom or a normal portage atom that contains the atom to match against
    @type	dbapi: portage.dbapi.dbapi
    @param	dbapi: one of the portage databases to use as information source
    @type	match_type: string
    @param	match_type: if != "default" passed as first argument to dbapi.xmatch
                            to apply the wanted visibility filters

    @rtype:		list of strings
    @return:	a list with the matching versions
    """
    if atom[2] == "~":
        return revisionMatch(atom, dbapi, match_type=match_type)
    if match_type == "default" or not hasattr(dbapi, "xmatch"):
        return dbapi.match(atom)
    return dbapi.xmatch(match_type, atom)


def revisionMatch(revisionAtom, dbapi, match_type="default"):
    """
    handler for the special >~, >=~, <=~ and <~ atoms that are supposed to behave
    as > and < except that they are limited to the same version, the range only
    applies to the revision part.

    @type	revisionAtom: string
    @param	revisionAtom: a <~ or >~ atom that contains the atom to match against
    @type	dbapi: portage.dbapi.dbapi
    @param	dbapi: one of the portage databases to use as information source
    @type	match_type: string
    @param	match_type: if != "default" passed as first argument to portdb.xmatch
                            to apply the wanted visibility filters

    @rtype:		list of strings
    @return:	a list with the matching versions
    """
    if match_type == "default" or not hasattr(dbapi, "xmatch"):
        if ":" in revisionAtom:
            mylist = dbapi.match(re.sub(r"-r[0-9]+(:[^ ]+)?$", r"\1", revisionAtom[2:]))
        else:
            mylist = dbapi.match(re.sub("-r[0-9]+$", "", revisionAtom[2:]))
    else:
        if ":" in revisionAtom:
            mylist = dbapi.xmatch(
                match_type, re.sub(r"-r[0-9]+(:[^ ]+)?$", r"\1", revisionAtom[2:])
            )
        else:
            mylist = dbapi.xmatch(match_type, re.sub("-r[0-9]+$", "", revisionAtom[2:]))
    rValue = []
    for v in mylist:
        r1 = pkgsplit(v)[-1][1:]
        r2 = pkgsplit(revisionAtom[3:])[-1][1:]
        if eval(r1 + " " + revisionAtom[0:2] + " " + r2):
            rValue.append(v)
    return rValue


def getMinUpgrade(vulnerableList, unaffectedList, portdbapi, vardbapi, minimize=True):
    """
    Checks if the systemstate is matching an atom in
    I{vulnerableList} and returns string describing
    the lowest version for the package that matches an atom in
    I{unaffectedList} and is greater than the currently installed
    version. It will return an empty list if the system is affected,
    and no upgrade is possible or None if the system is not affected.
    Both I{vulnerableList} and I{unaffectedList} should have the
    same base package.

    @type	vulnerableList: List of Strings
    @param	vulnerableList: atoms matching vulnerable package versions
    @type	unaffectedList: List of Strings
    @param	unaffectedList: atoms matching unaffected package versions
    @type	portdbapi:	portage.dbapi.porttree.portdbapi
    @param	portdbapi:	Ebuild repository
    @type	vardbapi:	portage.dbapi.vartree.vardbapi
    @param	vardbapi:	Installed package repository
    @type	minimize:	Boolean
    @param	minimize:	True for a least-change upgrade, False for emerge-like algorithm

    @rtype:		String | None
    @return:	the lowest unaffected version that is greater than
                            the installed version.
    """
    rValue = ""
    v_installed = reduce(operator.add, [match(v, vardbapi) for v in vulnerableList], [])
    u_installed = reduce(operator.add, [match(u, vardbapi) for u in unaffectedList], [])

    # remove all unaffected atoms from vulnerable list
    v_installed = list(set(v_installed).difference(set(u_installed)))

    if not v_installed:
        return None

    # this tuple holds all vulnerable atoms, and the related upgrade atom
    vuln_update = []
    avail_updates = set()
    for u in unaffectedList:
        # TODO: This had match_type="match-all" before. I don't think it should
        # since we disregarded masked items later anyway (match(=rValue, "porttree"))
        avail_updates.update(match(u, portdbapi))
    # if an atom is already installed, we should not consider it for upgrades
    avail_updates.difference_update(u_installed)

    for vuln in v_installed:
        update = ""
        for c in avail_updates:
            c_pv = portage.catpkgsplit(c)
            if (
                vercmp(c.version, vuln.version) > 0
                and (
                    update == "" or (minimize ^ (vercmp(c.version, update.version) > 0))
                )
                and portdbapi._pkg_str(c, None).slot
                == vardbapi._pkg_str(vuln, None).slot
            ):
                update = f"{c_pv[0]}/{c_pv[1]}-{c_pv[2]}"
                if c_pv[3] != "r0":  # we don't like -r0 for display
                    update = f"{update}-{c_pv[3]}"
                update = portdbapi._pkg_str(update, None)
        vuln_update.append([vuln, update])

    return vuln_update


def format_date(datestr):
    """
    Takes a date (announced, revised) date from a GLSA and formats
    it as readable text (i.e. "January 1, 2008").

    @type	date: String
    @param	date: the date string to reformat
    @rtype:		String
    @return:	a reformatted string, or the original string
                            if it cannot be reformatted.
    """
    splitdate = datestr.split("-", 2)
    if len(splitdate) != 3:
        return datestr

    # This cannot raise an error as we use () instead of []
    splitdate = (int(x) for x in splitdate)

    from datetime import date

    try:
        d = date(*splitdate)
    except ValueError:
        return datestr

    # TODO We could format to local date format '%x' here?
    return _unicode_decode(
        d.strftime("%B %d, %Y"), encoding=_encodings["content"], errors="replace"
    )


# simple Exception classes to catch specific errors


class GlsaTypeException(Exception):
    def __init__(self, doctype):
        Exception.__init__(self, f"wrong DOCTYPE: {doctype}")


class GlsaFormatException(Exception):
    pass


class GlsaArgumentException(Exception):
    pass


# GLSA xml data wrapper class


class Glsa:
    """
    This class is a wrapper for the XML data and provides methods to access
    and display the contained data.
    """

    def __init__(self, myid, myconfig, vardbapi, portdbapi):
        """
        Simple constructor to set the ID, store the config and gets the
        XML data by calling C{self.read()}.

        @type	myid: String
        @param	myid: String describing the id for the GLSA object (standard
                                  GLSAs have an ID of the form YYYYMM-nn) or an existing
                                  filename containing a GLSA.
        @type	myconfig: portage.config
        @param	myconfig: the config that should be used for this object.
        @type	vardbapi: portage.dbapi.vartree.vardbapi
        @param	vardbapi: installed package repository
        @type	portdbapi: portage.dbapi.porttree.portdbapi
        @param	portdbapi: ebuild repository
        """
        myid = _unicode_decode(myid, encoding=_encodings["content"], errors="strict")
        if re.match(r"\d{6}-\d{2}", myid):
            self.type = "id"
        elif os.path.exists(myid):
            self.type = "file"
        else:
            raise GlsaArgumentException(
                _(f"Given ID {myid} isn't a valid GLSA ID or filename.")
            )
        self.nr = myid
        self.config = myconfig
        self.vardbapi = vardbapi
        self.portdbapi = portdbapi
        self.read()

    def read(self):
        """
        Here we build the filename from the config and the ID and pass
        it to urllib to fetch it from the filesystem or a remote server.

        @rtype:		None
        @return:	None
        """
        if "GLSA_DIR" in self.config:
            repository = f"file://{self.config['GLSA_DIR']}/"
        else:
            repository = f"file://{self.config['PORTDIR']}/metadata/glsa/"
        if self.type == "file":
            myurl = f"file://{self.nr}"
        else:
            myurl = f"{repository}glsa-{self.nr}.xml"

        f = urllib_request_urlopen(myurl)
        try:
            self.parse(f)
        finally:
            f.close()

    def parse(self, myfile):
        """
        This method parses the XML file and sets up the internal data
        structures by calling the different helper functions in this
        module.

        @type	myfile: String
        @param	myfile: Filename to grab the XML data from
        @rtype:		None
        @return:	None
        """
        self.DOM = xml.dom.minidom.parse(myfile)
        if not self.DOM.doctype:
            raise GlsaTypeException(None)
        elif self.DOM.doctype.systemId == "http://www.gentoo.org/dtd/glsa.dtd":
            self.dtdversion = 0
        elif self.DOM.doctype.systemId == "http://www.gentoo.org/dtd/glsa-2.dtd":
            self.dtdversion = 2
        else:
            raise GlsaTypeException(self.DOM.doctype.systemId)
        myroot = self.DOM.getElementsByTagName("glsa")[0]
        if self.type == "id" and myroot.getAttribute("id") != self.nr:
            raise GlsaFormatException(
                _(
                    f"filename and internal id don't match: {myroot.getAttribute('id')} != {self.nr}"
                )
            )

        # the simple (single, required, top-level, #PCDATA) tags first
        self.title = getText(myroot.getElementsByTagName("title")[0], format="strip")
        self.synopsis = getText(
            myroot.getElementsByTagName("synopsis")[0], format="strip"
        )
        self.announced = format_date(
            getText(myroot.getElementsByTagName("announced")[0], format="strip")
        )

        # Support only format defined in GLSA DTD
        # <revised count="2">2007-12-30</revised>
        revisedEl = myroot.getElementsByTagName("revised")[0]
        self.revised = getText(revisedEl, format="strip")
        count = revisedEl.getAttribute("count")
        if not count:
            raise GlsaFormatException(
                f"Count attribute is missing or blank in GLSA: {myroot.getAttribute('id')}"
            )

        try:
            self.count = int(count)
        except ValueError:
            raise GlsaFormatException(
                f"Revision attribute in GLSA: {myroot.getAttribute('id')} is not an integer"
            )

        self.revised = format_date(self.revised)

        # now the optional and 0-n toplevel, #PCDATA tags and references
        try:
            self.access = getText(
                myroot.getElementsByTagName("access")[0], format="strip"
            )
        except IndexError:
            self.access = ""
        self.bugs = getMultiTagsText(myroot, "bug", format="strip")
        self.references = getMultiTagsText(
            myroot.getElementsByTagName("references")[0], "uri", format="keep"
        )

        # and now the formatted text elements
        self.description = getText(
            myroot.getElementsByTagName("description")[0], format="xml"
        )
        self.workaround = getText(
            myroot.getElementsByTagName("workaround")[0], format="xml"
        )
        self.resolution = getText(
            myroot.getElementsByTagName("resolution")[0], format="xml"
        )
        self.impact_text = getText(
            myroot.getElementsByTagName("impact")[0], format="xml"
        )
        self.impact_type = myroot.getElementsByTagName("impact")[0].getAttribute("type")
        try:
            self.background = getText(
                myroot.getElementsByTagName("background")[0], format="xml"
            )
        except IndexError:
            self.background = ""

        # finally the interesting tags (product, affected, package)
        self.glsatype = myroot.getElementsByTagName("product")[0].getAttribute("type")
        self.product = getText(
            myroot.getElementsByTagName("product")[0], format="strip"
        )
        self.affected = myroot.getElementsByTagName("affected")[0]
        self.packages = {}
        for p in self.affected.getElementsByTagName("package"):
            name = p.getAttribute("name")
            try:
                name = portage.dep.Atom(name)
            except portage.exception.InvalidAtom:
                raise GlsaFormatException(_(f"invalid package name: {name}"))
            if name != name.cp:
                raise GlsaFormatException(_(f"invalid package name: {name}"))
            name = name.cp
            if name not in self.packages:
                self.packages[name] = []
            tmp = {}
            tmp["arch"] = p.getAttribute("arch")
            tmp["auto"] = p.getAttribute("auto") == "yes"
            tmp["vul_vers"] = [
                makeVersion(v) for v in p.getElementsByTagName("vulnerable")
            ]
            tmp["unaff_vers"] = [
                makeVersion(v) for v in p.getElementsByTagName("unaffected")
            ]
            tmp["vul_atoms"] = [
                makeAtom(name, v) for v in p.getElementsByTagName("vulnerable")
            ]
            tmp["unaff_atoms"] = [
                makeAtom(name, v) for v in p.getElementsByTagName("unaffected")
            ]
            self.packages[name].append(tmp)
        # TODO: services aren't really used yet
        self.services = self.affected.getElementsByTagName("service")

    def dump(self, outstream=sys.stdout, encoding="utf-8"):
        """
        Dumps a plaintext representation of this GLSA to I{outfile} or
        B{stdout} if it is ommitted. You can specify an alternate
        I{encoding} if needed (default is utf-8).

        @type	outstream: File
        @param	outfile: Stream that should be used for writing
                                         (defaults to sys.stdout)
        """
        outstream = getattr(outstream, "buffer", outstream)
        outstream = codecs.getwriter(encoding)(outstream)
        width = 76
        buffer = "\n".join(
            (
                f"GLSA {self.nr}: ",
                f"{self.title}".center(width),
                "=" * width,
                wrap(self.synopsis, width, caption=_("Synopsis:         ")),
                _(f"Announced on:      {self.announced}"),
                _(f"Last revised on:   {self.revised} : %{self.count}\n"),
            )
        )
        outstream.write(buffer)
        if self.glsatype == "ebuild":
            for k in self.packages:
                pkg = self.packages[k]
                for path in pkg:
                    vul_vers = ", ".join(path["vul_vers"])
                    unaff_vers = ", ".join(path["unaff_vers"])
                    outstream.write(_(f"Affected package:  {k}\n"))
                    outstream.write(_("Affected archs:    "))
                    if path["arch"] == "*":
                        outstream.write(_("All\n"))
                    else:
                        outstream.write(f"{path['arch']}\n")
                    outstream.write(_(f"Vulnerable:        {vul_vers}\n"))
                    outstream.write(_(f"Unaffected:        {unaff_vers}\n\n"))
        elif self.glsatype == "infrastructure":
            pass
        if len(self.bugs) > 0:
            outstream.write(_(f"\nRelated bugs:      {', '.join(self.bugs)}\n"))
        if self.background:
            bg = wrap(self.background, width, caption=_("Background:       "))
            outstream.write(f"\n{bg}")
        myreferences = " ".join(
            r.replace(" ", SPACE_ESCAPE) + NEWLINE_ESCAPE for r in self.references
        )
        buffer = "\n".join(
            (
                wrap(self.description, width, caption=_("Description:      ")),
                wrap(self.impact_text, width, caption=_("Impact:           ")),
                wrap(self.workaround, width, caption=_("Workaround:       ")),
                wrap(self.resolution, width, caption=_("Resolution:       ")),
                wrap(myreferences, width, caption=_("References:       ")),
            )
        )
        outstream.write(f"\n{buffer}\n")

    def isVulnerable(self):
        """
        Tests if the system is affected by this GLSA by checking if any
        vulnerable package versions are installed. Also checks for affected
        architectures.

        @rtype:		Boolean
        @return:	True if the system is affected, False if not
        """
        rValue = False
        for k in self.packages:
            pkg = self.packages[k]
            for path in pkg:
                if path["arch"] == "*" or self.config["ARCH"] in path["arch"].split():
                    for v in path["vul_atoms"]:
                        rValue = rValue or (
                            len(match(v, self.vardbapi)) > 0
                            and None
                            != getMinUpgrade(
                                path["vul_atoms"],
                                path["unaff_atoms"],
                                self.portdbapi,
                                self.vardbapi,
                            )
                        )
        return rValue

    def isInjected(self):
        """
        Looks if the GLSA ID is in the GLSA checkfile to check if this
        GLSA should be marked as applied.

        @rtype:		Boolean
        @returns:	True if the GLSA is in the inject file, False if not
        """
        if not os.access(
            os.path.join(self.config["EROOT"], PRIVATE_PATH, "glsa_injected"), os.R_OK
        ):
            return False
        return self.nr in get_applied_glsas(self.config)

    def inject(self):
        """
        Puts the ID of this GLSA into the GLSA checkfile, so it won't
        show up on future checks. Should be called after a GLSA is
        applied or on explicit user request.

        @rtype:		None
        @return:	None
        """
        if not self.isInjected():
            checkfile = io.open(
                _unicode_encode(
                    os.path.join(self.config["EROOT"], PRIVATE_PATH, "glsa_injected"),
                    encoding=_encodings["fs"],
                    errors="strict",
                ),
                mode="a+",
                encoding=_encodings["content"],
                errors="strict",
            )
            checkfile.write(_unicode_decode(self.nr + "\n"))
            checkfile.close()

    def getMergeList(self, least_change=True):
        """
        Returns the list of package-versions that have to be merged to
        apply this GLSA properly. The versions are as low as possible
        while avoiding downgrades (see L{getMinUpgrade}).

        @type	least_change: Boolean
        @param	least_change: True if the smallest possible upgrade should be selected,
                                False for an emerge-like algorithm
        @rtype:		List of Strings
        @return:	list of package-versions that have to be merged
        """
        return list(
            set(
                update
                for (vuln, update) in self.getAffectionTable(least_change)
                if update
            )
        )

    def getAffectionTable(self, least_change=True):
        """
        Will initialize the self.systemAffection list of
        atoms installed on the system that are affected
        by this GLSA, and the atoms that are minimal upgrades.
        """
        systemAffection = []
        for pkg in self.packages.keys():
            for path in self.packages[pkg]:
                update = getMinUpgrade(
                    path["vul_atoms"],
                    path["unaff_atoms"],
                    self.portdbapi,
                    self.vardbapi,
                    minimize=least_change,
                )
                if update:
                    systemAffection.extend(update)
        return systemAffection



# Copyright 2013-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import itertools

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class OrChoicesTestCase(TestCase):
    def testOrChoices(self):
        ebuilds = {
            "dev-lang/vala-0.20.0": {"EAPI": "5", "SLOT": "0.20"},
            "dev-lang/vala-0.18.0": {"EAPI": "5", "SLOT": "0.18"},
            # "dev-libs/gobject-introspection-1.36.0" : {
            # 	"EAPI": "5",
            # 	"RDEPEND" : "!<dev-lang/vala-0.20.0",
            # },
            "dev-libs/gobject-introspection-1.34.0": {"EAPI": "5"},
            "sys-apps/systemd-ui-2": {
                "EAPI": "5",
                "RDEPEND": "|| ( dev-lang/vala:0.20 dev-lang/vala:0.18 )",
            },
        }

        installed = {
            "dev-lang/vala-0.18.0": {"EAPI": "5", "SLOT": "0.18"},
            "dev-libs/gobject-introspection-1.34.0": {"EAPI": "5"},
            "sys-apps/systemd-ui-2": {
                "EAPI": "5",
                "RDEPEND": "|| ( dev-lang/vala:0.20 dev-lang/vala:0.18 )",
            },
        }

        world = ["dev-libs/gobject-introspection", "sys-apps/systemd-ui"]

        test_cases = (
            # Demonstrate that vala:0.20 update is pulled in, for bug #478188
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True},
                success=True,
                all_permutations=True,
                mergelist=["dev-lang/vala-0.20.0"],
            ),
            # Verify that vala:0.20 is not pulled in without --deep
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True},
                success=True,
                all_permutations=True,
                mergelist=[],
            ),
            # Verify that vala:0.20 is not pulled in without --update
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--selective": True, "--deep": True},
                success=True,
                all_permutations=True,
                mergelist=[],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def testInitiallyUnsatisfied(self):

        ebuilds = {
            "app-misc/A-1": {"EAPI": "5", "SLOT": "0/1"},
            "app-misc/A-2": {"EAPI": "5", "SLOT": "0/2"},
            "app-misc/B-0": {"EAPI": "5", "RDEPEND": "app-misc/A:="},
            "app-misc/C-0": {"EAPI": "5", "RDEPEND": "|| ( app-misc/X <app-misc/A-2 )"},
        }

        installed = {
            "app-misc/A-1": {"EAPI": "5", "SLOT": "0/1"},
            "app-misc/B-0": {"EAPI": "5", "RDEPEND": "app-misc/A:0/1="},
            "app-misc/C-0": {"EAPI": "5", "RDEPEND": "|| ( app-misc/X <app-misc/A-2 )"},
        }

        world = ["app-misc/B", "app-misc/C"]

        test_cases = (
            # Test bug #522652, where the unsatisfiable app-misc/X
            # atom is selected, and the dependency is placed into
            # _initially_unsatisfied_deps where it is ignored, causing
            # upgrade to app-misc/A-2 (breaking a dependency of
            # app-misc/C-0).
            ResolverPlaygroundTestCase(
                ["app-misc/A"], options={}, success=True, mergelist=["app-misc/A-1"]
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world, debug=False
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def testUseMask(self):

        profile = {
            "use.mask": ("abi_ppc_32",),
        }

        ebuilds = {
            "sys-libs/A-1": {
                "EAPI": "5",
                "RDEPEND": "|| ( sys-libs/zlib[abi_ppc_32(-)] "
                + "sys-libs/zlib[abi_x86_32(-)] )",
            },
            "sys-libs/zlib-1.2.8-r1": {"EAPI": "5", "IUSE": "abi_ppc_32 abi_x86_32"},
            "sys-libs/zlib-1.2.8": {"EAPI": "5", "IUSE": ""},
        }

        test_cases = (
            # bug #515584: We want to prefer choices that do
            # not require changes to use.mask or use.force.
            # In this case, abi_ppc_32 is use.masked in the
            # profile, so we want to avoid that choice.
            ResolverPlaygroundTestCase(
                ["sys-libs/A"],
                options={},
                success=False,
                use_changes={"sys-libs/zlib-1.2.8-r1": {"abi_x86_32": True}},
                mergelist=["sys-libs/zlib-1.2.8-r1", "sys-libs/A-1"],
            ),
        )

        playground = ResolverPlayground(ebuilds=ebuilds, profile=profile, debug=False)
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def testConflictMissedUpdate(self):

        ebuilds = {
            "dev-lang/ocaml-4.02.1": {
                "EAPI": "5",
                "SLOT": "0/4.02.1",
            },
            "dev-lang/ocaml-4.01.0": {
                "EAPI": "5",
                "SLOT": "0/4.01.0",
            },
            "dev-ml/lablgl-1.05": {
                "EAPI": "5",
                "DEPEND": (
                    ">=dev-lang/ocaml-3.10.2:= "
                    "|| ( dev-ml/labltk:= <dev-lang/ocaml-4.02 )"
                ),
                "RDEPEND": (
                    ">=dev-lang/ocaml-3.10.2:= "
                    "|| ( dev-ml/labltk:= <dev-lang/ocaml-4.02 )"
                ),
            },
            "dev-ml/labltk-8.06.0": {
                "EAPI": "5",
                "SLOT": "0/8.06.0",
                "DEPEND": ">=dev-lang/ocaml-4.02:=",
                "RDEPEND": ">=dev-lang/ocaml-4.02:=",
            },
        }

        installed = {
            "dev-lang/ocaml-4.01.0": {
                "EAPI": "5",
                "SLOT": "0/4.01.0",
            },
            "dev-ml/lablgl-1.05": {
                "EAPI": "5",
                "DEPEND": (
                    ">=dev-lang/ocaml-3.10.2:0/4.01.0= "
                    "|| ( dev-ml/labltk:= <dev-lang/ocaml-4.02 )"
                ),
                "RDEPEND": (
                    ">=dev-lang/ocaml-3.10.2:0/4.01.0= "
                    "|| ( dev-ml/labltk:= <dev-lang/ocaml-4.02 )"
                ),
            },
        }

        world = (
            "dev-lang/ocaml",
            "dev-ml/lablgl",
        )

        test_cases = (
            # bug #531656: If an ocaml update is desirable,
            # then we need to pull in dev-ml/labltk.
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True},
                success=True,
                mergelist=[
                    "dev-lang/ocaml-4.02.1",
                    "dev-ml/labltk-8.06.0",
                    "dev-ml/lablgl-1.05",
                ],
            ),
        )

        playground = ResolverPlayground(
            debug=False, ebuilds=ebuilds, installed=installed, world=world
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            # Disable debug so that cleanup works.
            playground.debug = False
            playground.cleanup()

    def test_python_slot(self):
        ebuilds = {
            "dev-lang/python-3.8": {"EAPI": "7", "SLOT": "3.8"},
            "dev-lang/python-3.7": {"EAPI": "7", "SLOT": "3.7"},
            "dev-lang/python-3.6": {"EAPI": "7", "SLOT": "3.6"},
            "app-misc/bar-1": {
                "EAPI": "7",
                "IUSE": "python_targets_python3_6 +python_targets_python3_7",
                "RDEPEND": "python_targets_python3_7? ( dev-lang/python:3.7 ) python_targets_python3_6? ( dev-lang/python:3.6 )",
            },
            "app-misc/foo-1": {
                "EAPI": "7",
                "RDEPEND": "|| ( dev-lang/python:3.8 dev-lang/python:3.7 dev-lang/python:3.6 )",
            },
        }

        installed = {
            "dev-lang/python-3.7": {"EAPI": "7", "SLOT": "3.7"},
            "app-misc/bar-1": {
                "EAPI": "7",
                "IUSE": "python_targets_python3_6 +python_targets_python3_7",
                "USE": "python_targets_python3_7",
                "RDEPEND": "dev-lang/python:3.7",
            },
            "app-misc/foo-1": {
                "EAPI": "7",
                "RDEPEND": "|| ( dev-lang/python:3.8 dev-lang/python:3.7 dev-lang/python:3.6 )",
            },
        }

        world = ["app-misc/foo", "app-misc/bar"]

        test_cases = (
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True},
                success=True,
                mergelist=["dev-lang/python-3.8"],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world, debug=False
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.debug = False
            playground.cleanup()

        installed = {
            "dev-lang/python-3.8": {"EAPI": "7", "SLOT": "3.8"},
            "dev-lang/python-3.7": {"EAPI": "7", "SLOT": "3.7"},
            "app-misc/bar-1": {
                "EAPI": "7",
                "IUSE": "python_targets_python3_6 +python_targets_python3_7",
                "USE": "python_targets_python3_7",
                "RDEPEND": "dev-lang/python:3.7",
            },
            "app-misc/foo-1": {
                "EAPI": "7",
                "RDEPEND": "|| ( dev-lang/python:3.8 dev-lang/python:3.7 dev-lang/python:3.6 )",
            },
        }

        test_cases = (
            # Test for bug 707108, where a new python slot was erroneously
            # removed by emerge --depclean.
            ResolverPlaygroundTestCase(
                [],
                options={"--depclean": True},
                success=True,
                cleanlist=[],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world, debug=False
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.debug = False
            playground.cleanup()

    def test_virtual_w3m(self):
        ebuilds = {
            "app-text/xmlto-0.0.28-r1": {
                "EAPI": "7",
                "RDEPEND": "|| ( virtual/w3m www-client/lynx www-client/elinks )",
            },
            "www-client/elinks-0.13_pre_pre20180225": {
                "EAPI": "7",
            },
            "www-client/lynx-2.9.0_pre4": {
                "EAPI": "7",
            },
            "virtual/w3m-0": {
                "EAPI": "7",
                "RDEPEND": "|| ( www-client/w3m www-client/w3mmee )",
            },
            "www-client/w3m-0.5.3_p20190105": {
                "EAPI": "7",
            },
            "www-client/w3mmee-0.3.2_p24-r10": {
                "EAPI": "7",
            },
        }

        installed = {
            "app-text/xmlto-0.0.28-r1": {
                "EAPI": "7",
                "RDEPEND": "|| ( virtual/w3m www-client/lynx www-client/elinks )",
            },
            "www-client/elinks-0.13_pre_pre20180225": {
                "EAPI": "7",
            },
            "www-client/lynx-2.9.0_pre4": {
                "EAPI": "7",
            },
        }

        world = ["app-text/xmlto", "www-client/elinks", "www-client/lynx"]

        test_cases = (
            # Test for bug 649622 (without www-client/w3m installed),
            # where virtual/w3m was pulled in only to be removed by the
            # next emerge --depclean.
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True},
                success=True,
                mergelist=[],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world, debug=False
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.debug = False
            playground.cleanup()

        installed = dict(
            itertools.chain(
                installed.items(),
                {
                    "www-client/w3m-0.5.3_p20190105": {
                        "EAPI": "7",
                    },
                }.items(),
            )
        )

        test_cases = (
            # Test for bug 649622 (with www-client/w3m installed),
            # where virtual/w3m was pulled in only to be removed by the
            # next emerge --depclean.
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True},
                success=True,
                mergelist=[],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world, debug=False
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.debug = False
            playground.cleanup()

        installed = dict(
            itertools.chain(
                installed.items(),
                {
                    "virtual/w3m-0": {
                        "EAPI": "7",
                        "RDEPEND": "|| ( www-client/w3m www-client/w3mmee )",
                    },
                }.items(),
            )
        )

        test_cases = (
            # Test for bug 649622, where virtual/w3m is removed by
            # emerge --depclean immediately after it's installed
            # by a world update. Note that removal of virtual/w3m here
            # is essentially indistinguishable from removal of
            # dev-util/cmake-bootstrap in the depclean test case for
            # bug 703440.
            ResolverPlaygroundTestCase(
                [],
                options={"--depclean": True},
                success=True,
                cleanlist=["virtual/w3m-0", "www-client/w3m-0.5.3_p20190105"],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world, debug=False
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.debug = False
            playground.cleanup()

        test_cases = (
            # Test for behavior reported in bug 649622 comment #10, where
            # depclean removed virtual/w3m-0 even though www-client/w3m
            # was in the world file. Since nothing is removed here, it
            # means that we have not reproduced the behavior reported in
            # this comment.
            ResolverPlaygroundTestCase(
                [],
                options={"--depclean": True},
                success=True,
                cleanlist=[],
            ),
        )

        world += ["www-client/w3m"]

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world, debug=False
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.debug = False
            playground.cleanup()

    def test_virtual_w3m_realistic(self):
        """
        Test for bug 649622 with realistic www-client/w3m dependencies copied
        from real ebuilds.
        """
        ebuilds = {
            "app-misc/neofetch-6.1.0": {"EAPI": "7", "RDEPEND": "www-client/w3m"},
            "app-text/xmlto-0.0.28-r1": {
                "EAPI": "7",
                "RDEPEND": "|| ( virtual/w3m www-client/lynx www-client/elinks )",
            },
            "mail-client/neomutt-20191207": {
                "EAPI": "7",
                "RDEPEND": "|| ( www-client/lynx www-client/w3m www-client/elinks )",
            },
            "www-client/elinks-0.13_pre_pre20180225": {
                "EAPI": "7",
            },
            "www-client/lynx-2.9.0_pre4": {
                "EAPI": "7",
            },
            "virtual/w3m-0": {
                "EAPI": "7",
                "RDEPEND": "|| ( www-client/w3m www-client/w3mmee )",
            },
            "www-client/w3m-0.5.3_p20190105": {
                "EAPI": "7",
            },
            "www-client/w3mmee-0.3.2_p24-r10": {
                "EAPI": "7",
            },
            "x11-base/xorg-server-1.20.7": {
                "EAPI": "7",
                "RDEPEND": "|| ( www-client/links www-client/lynx www-client/w3m ) app-text/xmlto",
            },
        }

        installed = {
            "app-misc/neofetch-6.1.0": {"EAPI": "7", "RDEPEND": "www-client/w3m"},
            "app-text/xmlto-0.0.28-r1": {
                "EAPI": "7",
                "RDEPEND": "|| ( virtual/w3m www-client/lynx www-client/elinks )",
            },
            "mail-client/neomutt-20191207": {
                "EAPI": "7",
                "RDEPEND": "|| ( www-client/lynx www-client/w3m www-client/elinks )",
            },
            "www-client/lynx-2.9.0_pre4": {
                "EAPI": "7",
            },
            "www-client/w3m-0.5.3_p20190105": {
                "EAPI": "7",
            },
            "x11-base/xorg-server-1.20.7": {
                "EAPI": "7",
                "RDEPEND": "|| ( www-client/links www-client/lynx www-client/w3m ) app-text/xmlto",
            },
        }

        world = [
            "app-misc/neofetch",
            "mail-client/neomutt",
            "www-client/lynx",
            "x11-base/xorg-server",
        ]

        test_cases = (
            # Test for bug 649622 (with www-client/w3m installed via
            # xorg-server dependency), where virtual/w3m was pulled in
            # only to be removed by the next emerge --depclean. Note
            # that graph_order must be deterministic in order to achieve
            # deterministic results which are consistent between both
            # update and removal (depclean) actions.
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True},
                success=True,
                mergelist=["virtual/w3m-0"],
                graph_order=[
                    "@world",
                    "@profile",
                    "@selected",
                    "@system",
                    "[nomerge]app-misc/neofetch-6.1.0",
                    "[nomerge]mail-client/neomutt-20191207",
                    "[nomerge]www-client/lynx-2.9.0_pre4",
                    "[nomerge]x11-base/xorg-server-1.20.7",
                    "[nomerge]app-text/xmlto-0.0.28-r1",
                    "[nomerge]www-client/w3m-0.5.3_p20190105",
                    "virtual/w3m-0",
                ],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world, debug=False
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.debug = False
            playground.cleanup()

        installed = dict(
            itertools.chain(
                installed.items(),
                {
                    "virtual/w3m-0": {
                        "EAPI": "7",
                        "RDEPEND": "|| ( www-client/w3m www-client/w3mmee )",
                    },
                }.items(),
            )
        )

        test_cases = (
            # Test for bug 649622, where virtual/w3m is removed by
            # emerge --depclean immediately after it's installed
            # by a world update. Since virtual/w3m-0 is not removed
            # here, this case fails to reproduce bug 649622. Note
            # that graph_order must be deterministic in order to achieve
            # deterministic results which are consistent between both
            # update and removal (depclean) actions.
            ResolverPlaygroundTestCase(
                [],
                options={"--depclean": True},
                success=True,
                cleanlist=[],
                graph_order=[
                    "@world",
                    "@____depclean_protected_set____",
                    "@profile",
                    "@selected",
                    "@system",
                    "[nomerge]app-misc/neofetch-6.1.0",
                    "[nomerge]mail-client/neomutt-20191207",
                    "[nomerge]www-client/lynx-2.9.0_pre4",
                    "[nomerge]x11-base/xorg-server-1.20.7",
                    "[nomerge]app-text/xmlto-0.0.28-r1",
                    "[nomerge]www-client/w3m-0.5.3_p20190105",
                    "[nomerge]virtual/w3m-0",
                ],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world, debug=False
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.debug = False
            playground.cleanup()


class OrChoicesLibpostprocTestCase(TestCase):
    def testOrChoicesLibpostproc(self):
        # This test case is expected to fail after the fix for bug 706278,
        # since the "undesirable" slot upgrade which triggers a blocker conflict
        # in this test case is practically indistinguishable from a desirable
        # slot upgrade. This particular blocker conflict is no longer relevant,
        # since current versions of media-libs/libpostproc are no longer
        # compatible with any available media-video/ffmpeg slot. In order to
        # solve this test case, some fancy backtracking (like for bug 382421)
        # will be required.
        self.todo = True

        ebuilds = {
            "media-video/ffmpeg-0.10": {"EAPI": "5", "SLOT": "0.10"},
            "media-video/ffmpeg-1.2.2": {"EAPI": "5", "SLOT": "0"},
            "media-libs/libpostproc-0.8.0.20121125": {"EAPI": "5"},
            "media-plugins/gst-plugins-ffmpeg-0.10.13_p201211-r1": {
                "EAPI": "5",
                "RDEPEND": "|| ( media-video/ffmpeg:0 media-libs/libpostproc )",
            },
        }

        installed = {
            "media-video/ffmpeg-0.10": {"EAPI": "5", "SLOT": "0.10"},
            "media-libs/libpostproc-0.8.0.20121125": {"EAPI": "5"},
            "media-plugins/gst-plugins-ffmpeg-0.10.13_p201211-r1": {
                "EAPI": "5",
                "RDEPEND": "|| ( media-video/ffmpeg:0 media-libs/libpostproc )",
            },
        }

        world = ["media-plugins/gst-plugins-ffmpeg"]

        test_cases = (
            # Demonstrate that libpostproc is preferred
            # over ffmpeg:0 for bug #480736.
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True},
                success=True,
                all_permutations=True,
                mergelist=[],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world, debug=False
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.debug = False
            playground.cleanup()



# Copyright 2007-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import subprocess

from portage import os
from portage import _unicode_decode
from portage._sets.base import PackageSet
from portage._sets import SetConfigError

__all__ = ["CommandOutputSet"]


class CommandOutputSet(PackageSet):
    """This class creates a PackageSet from the output of a shell command.
    The shell command should produce one atom per line, that is:

    >>> atom1
        atom2
        ...
        atomN

    Args:
      name: A string that identifies the set.
      command: A string or sequence identifying the command to run
      (see the subprocess.Popen documentaion for the format)
    """

    _operations = ["merge", "unmerge"]

    def __init__(self, command):
        super(CommandOutputSet, self).__init__()
        self._command = command
        self.description = "Package set generated from output of '%s'" % self._command

    def load(self):
        pipe = subprocess.Popen(self._command, stdout=subprocess.PIPE, shell=True)
        stdout, stderr = pipe.communicate()
        if pipe.wait() == os.EX_OK:
            self._setAtoms(_unicode_decode(stdout).splitlines())

    def singleBuilder(self, options, settings, trees):
        if not "command" in options:
            raise SetConfigError("no command specified")
        return CommandOutputSet(options["command"])

    singleBuilder = classmethod(singleBuilder)



# testConfigFileSet.py -- Portage Unit Testing Functionality
# Copyright 2010 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.dep import Atom
from portage.exception import InvalidAtom
from portage.tests import TestCase
from portage._sets.base import InternalPackageSet


class InternalPackageSetTestCase(TestCase):
    """Simple Test Case for InternalPackageSet"""

    def testInternalPackageSet(self):
        i1_atoms = set(("dev-libs/A", ">=dev-libs/A-1", "dev-libs/B"))
        i2_atoms = set(("dev-libs/A", "dev-libs/*", "dev-libs/C"))

        i1 = InternalPackageSet(initial_atoms=i1_atoms)
        i2 = InternalPackageSet(initial_atoms=i2_atoms, allow_wildcard=True)
        self.assertRaises(InvalidAtom, InternalPackageSet, initial_atoms=i2_atoms)

        self.assertEqual(i1.getAtoms(), i1_atoms)
        self.assertEqual(i2.getAtoms(), i2_atoms)

        new_atom = Atom("*/*", allow_wildcard=True)
        self.assertRaises(InvalidAtom, i1.add, new_atom)
        i2.add(new_atom)

        i2_atoms.add(new_atom)

        self.assertEqual(i1.getAtoms(), i1_atoms)
        self.assertEqual(i2.getAtoms(), i2_atoms)

        removed_atom = Atom("dev-libs/A")

        i1.remove(removed_atom)
        i2.remove(removed_atom)

        i1_atoms.remove(removed_atom)
        i2_atoms.remove(removed_atom)

        self.assertEqual(i1.getAtoms(), i1_atoms)
        self.assertEqual(i2.getAtoms(), i2_atoms)

        update_atoms = [Atom("dev-libs/C"), Atom("dev-*/C", allow_wildcard=True)]

        self.assertRaises(InvalidAtom, i1.update, update_atoms)
        i2.update(update_atoms)

        i2_atoms.update(update_atoms)

        self.assertEqual(i1.getAtoms(), i1_atoms)
        self.assertEqual(i2.getAtoms(), i2_atoms)

        replace_atoms = [Atom("dev-libs/D"), Atom("*-libs/C", allow_wildcard=True)]

        self.assertRaises(InvalidAtom, i1.replace, replace_atoms)
        i2.replace(replace_atoms)

        i2_atoms = set(replace_atoms)

        self.assertEqual(i2.getAtoms(), i2_atoms)



# Copyright 2010-2013 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

__all__ = ["cacheddir", "listdir"]

import errno
import stat


from portage import os
from portage.const import VCS_DIRS
from portage.exception import DirectoryNotFound, PermissionDenied, PortageException
from portage.util import normalize_path

# The global dircache is no longer supported, since it could
# be a memory leak for API consumers. Any cacheddir callers
# should use higher-level caches instead, when necessary.
# TODO: Remove dircache variable after stable portage does
# not use is (keep it for now, in case API consumers clear
# it manually).
dircache = {}


def cacheddir(
    my_original_path, ignorecvs, ignorelist, EmptyOnError, followSymlinks=True
):
    mypath = normalize_path(my_original_path)
    try:
        pathstat = os.stat(mypath)
        if not stat.S_ISDIR(pathstat.st_mode):
            raise DirectoryNotFound(mypath)
    except EnvironmentError as e:
        if e.errno == PermissionDenied.errno:
            raise PermissionDenied(mypath)
        del e
        return [], []
    except PortageException:
        return [], []
    else:
        try:
            fpaths = os.listdir(mypath)
        except EnvironmentError as e:
            if e.errno != errno.EACCES:
                raise
            del e
            raise PermissionDenied(mypath)
        ftype = []
        for x in fpaths:
            try:
                if followSymlinks:
                    pathstat = os.stat(mypath + "/" + x)
                else:
                    pathstat = os.lstat(mypath + "/" + x)

                if stat.S_ISREG(pathstat[stat.ST_MODE]):
                    ftype.append(0)
                elif stat.S_ISDIR(pathstat[stat.ST_MODE]):
                    ftype.append(1)
                elif stat.S_ISLNK(pathstat[stat.ST_MODE]):
                    ftype.append(2)
                else:
                    ftype.append(3)
            except (IOError, OSError):
                ftype.append(3)

    if ignorelist or ignorecvs:
        ret_list = []
        ret_ftype = []
        for file_path, file_type in zip(fpaths, ftype):
            if file_path in ignorelist:
                pass
            elif ignorecvs:
                if file_path[:2] != ".#" and not (
                    file_type == 1 and file_path in VCS_DIRS
                ):
                    ret_list.append(file_path)
                    ret_ftype.append(file_type)
    else:
        ret_list = fpaths
        ret_ftype = ftype

    return ret_list, ret_ftype


def listdir(
    mypath,
    recursive=False,
    filesonly=False,
    ignorecvs=False,
    ignorelist=[],
    followSymlinks=True,
    EmptyOnError=False,
    dirsonly=False,
):
    """
    Portage-specific implementation of os.listdir

    @param mypath: Path whose contents you wish to list
    @type mypath: String
    @param recursive: Recursively scan directories contained within mypath
    @type recursive: Boolean
    @param filesonly; Only return files, not more directories
    @type filesonly: Boolean
    @param ignorecvs: Ignore VCS directories
    @type ignorecvs: Boolean
    @param ignorelist: List of filenames/directories to exclude
    @type ignorelist: List
    @param followSymlinks: Follow Symlink'd files and directories
    @type followSymlinks: Boolean
    @param EmptyOnError: Return [] if an error occurs (deprecated, always True)
    @type EmptyOnError: Boolean
    @param dirsonly: Only return directories.
    @type dirsonly: Boolean
    @rtype: List
    @return: A list of files and directories (or just files or just directories) or an empty list.
    """

    fpaths, ftype = cacheddir(
        mypath, ignorecvs, ignorelist, EmptyOnError, followSymlinks
    )

    if fpaths is None:
        fpaths = []
    if ftype is None:
        ftype = []

    if not (filesonly or dirsonly or recursive):
        return fpaths

    if recursive:
        stack = list(zip(fpaths, ftype))
        fpaths = []
        ftype = []
        while stack:
            file_path, file_type = stack.pop()
            fpaths.append(file_path)
            ftype.append(file_type)
            if file_type == 1:
                subdir_list, subdir_types = cacheddir(
                    os.path.join(mypath, file_path),
                    ignorecvs,
                    ignorelist,
                    EmptyOnError,
                    followSymlinks,
                )
                stack.extend(
                    (os.path.join(file_path, x), x_type)
                    for x, x_type in zip(subdir_list, subdir_types)
                )

    if filesonly:
        fpaths = [x for x, x_type in zip(fpaths, ftype) if x_type == 0]

    elif dirsonly:
        fpaths = [x for x, x_type in zip(fpaths, ftype) if x_type == 1]

    return fpaths



# Copyright 2015 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2
#
# This is a minimalistic derivation of Python's deprecated formatter module,
# supporting only the methods related to style, literal data, and line breaks.

import sys


class AbstractFormatter:
    """The standard formatter."""

    def __init__(self, writer):
        self.writer = writer  # Output device
        self.style_stack = []  # Other state, e.g. color
        self.hard_break = True  # Have a hard break

    def add_line_break(self):
        if not self.hard_break:
            self.writer.send_line_break()
        self.hard_break = True

    def add_literal_data(self, data):
        if not data:
            return
        self.hard_break = data[-1:] == "\n"
        self.writer.send_literal_data(data)

    def push_style(self, *styles):
        for style in styles:
            self.style_stack.append(style)
        self.writer.new_styles(tuple(self.style_stack))

    def pop_style(self, n=1):
        del self.style_stack[-n:]
        self.writer.new_styles(tuple(self.style_stack))


class NullWriter:
    """Minimal writer interface to use in testing & inheritance.

    A writer which only provides the interface definition; no actions are
    taken on any methods.  This should be the base class for all writers
    which do not need to inherit any implementation methods.
    """

    def __init__(self):
        pass

    def flush(self):
        pass

    def new_styles(self, styles):
        pass

    def send_line_break(self):
        pass

    def send_literal_data(self, data):
        pass


class DumbWriter(NullWriter):
    """Simple writer class which writes output on the file object passed in
    as the file parameter or, if file is omitted, on standard output.
    """

    def __init__(self, file=None, maxcol=None):
        NullWriter.__init__(self)
        self.file = file or sys.stdout

    def flush(self):
        self.file.flush()

    def send_line_break(self):
        self.file.write("\n")

    def send_literal_data(self, data):
        self.file.write(data)



# Copyright 2014 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import collections

from portage.dep import Atom
from portage.tests import TestCase
from _emerge.resolver.package_tracker import PackageTracker, PackageTrackerDbapiWrapper


class PackageTrackerTestCase(TestCase):

    FakePackage = collections.namedtuple(
        "FakePackage", ["root", "cp", "cpv", "slot", "slot_atom", "version", "repo"]
    )

    FakeConflict = collections.namedtuple(
        "FakeConflict", ["description", "root", "pkgs"]
    )

    def make_pkg(self, root, atom, repo="test_repo"):
        atom = Atom(atom)
        slot_atom = Atom("%s:%s" % (atom.cp, atom.slot))
        slot = atom.slot

        return self.FakePackage(
            root=root,
            cp=atom.cp,
            cpv=atom.cpv,
            slot=slot,
            slot_atom=slot_atom,
            version=atom.version,
            repo=repo,
        )

    def make_conflict(self, description, root, pkgs):
        return self.FakeConflict(description=description, root=root, pkgs=pkgs)

    def test_add_remove_discard(self):
        p = PackageTracker()

        x1 = self.make_pkg("/", "=dev-libs/X-1:0")
        x2 = self.make_pkg("/", "=dev-libs/X-2:0")

        p.add_pkg(x1)
        self.assertTrue(x1 in p)
        self.assertTrue(p.contains(x1, installed=True))
        self.assertTrue(p.contains(x1, installed=False))
        p.remove_pkg(x1)
        self.assertTrue(x1 not in p)

        p.add_pkg(x1)
        self.assertTrue(x1 in p)
        p.add_pkg(x1)
        self.assertTrue(x1 in p)

        self.assertRaises(KeyError, p.remove_pkg, x2)

        p.add_pkg(x2)
        self.assertTrue(x2 in p)
        p.remove_pkg(x2)
        self.assertTrue(x2 not in p)
        p.discard_pkg(x2)
        self.assertTrue(x2 not in p)
        p.add_pkg(x2)
        self.assertTrue(x2 in p)

        all_pkgs = list(p.all_pkgs("/"))
        self.assertEqual(len(all_pkgs), 2)
        self.assertTrue(all_pkgs[0] is x1 and all_pkgs[1] is x2)

        self.assertEqual(len(list(p.all_pkgs("/"))), 2)
        self.assertEqual(len(list(p.all_pkgs("/xxx"))), 0)

    def test_match(self):
        p = PackageTracker()
        x1 = self.make_pkg("/", "=dev-libs/X-1:0")
        x2 = self.make_pkg("/", "=dev-libs/X-2:0")
        x3 = self.make_pkg("/", "=dev-libs/X-3:1")

        p.add_pkg(x2)
        p.add_pkg(x1)

        matches = list(p.match("/", Atom("=dev-libs/X-1")))
        self.assertTrue(x1 in matches)
        self.assertEqual(len(matches), 1)

        matches = list(p.match("/", Atom("dev-libs/X")))
        self.assertTrue(x1 is matches[0] and x2 is matches[1])
        self.assertEqual(len(matches), 2)

        matches = list(p.match("/xxx", Atom("dev-libs/X")))
        self.assertEqual(len(matches), 0)

        matches = list(p.match("/", Atom("dev-libs/Y")))
        self.assertEqual(len(matches), 0)

        p.add_pkg(x3)
        matches = list(p.match("/", Atom("dev-libs/X")))
        self.assertTrue(x1 is matches[0] and x2 is matches[1] and x3 is matches[2])
        self.assertEqual(len(matches), 3)

        p.remove_pkg(x3)
        matches = list(p.match("/", Atom("dev-libs/X")))
        self.assertTrue(x1 is matches[0] and x2 is matches[1])
        self.assertEqual(len(matches), 2)

    def test_dbapi_interface(self):
        p = PackageTracker()
        dbapi = PackageTrackerDbapiWrapper("/", p)
        installed = self.make_pkg("/", "=dev-libs/X-0:0")
        x1 = self.make_pkg("/", "=dev-libs/X-1:0")
        x2 = self.make_pkg("/", "=dev-libs/X-2:0")
        x3 = self.make_pkg("/", "=dev-libs/X-3:0")
        x4 = self.make_pkg("/", "=dev-libs/X-4:6")
        x5 = self.make_pkg("/xxx", "=dev-libs/X-5:6")

        def check_dbapi(pkgs):
            all_pkgs = set(dbapi)
            self.assertEqual(len(all_pkgs), len(pkgs))

            x_atom = "dev-libs/X"
            y_atom = "dev-libs/Y"
            matches = dbapi.cp_list(x_atom)
            for pkg in pkgs:
                if pkg.root == "/" and pkg.cp == x_atom:
                    self.assertTrue(pkg in matches)
            self.assertTrue(not dbapi.cp_list(y_atom))
            matches = dbapi.match(Atom(x_atom))
            for pkg in pkgs:
                if pkg.root == "/" and pkg.cp == x_atom:
                    self.assertTrue(pkg in matches)
            self.assertTrue(not dbapi.match(Atom(y_atom)))

        check_dbapi([])

        p.add_installed_pkg(installed)
        check_dbapi([installed])

        p.add_pkg(x1)
        check_dbapi([x1])

        p.remove_pkg(x1)
        check_dbapi([installed])

        dbapi.cpv_inject(x1)
        check_dbapi([x1])

        dbapi.cpv_inject(x2)
        check_dbapi([x1, x2])

        p.remove_pkg(x1)
        check_dbapi([x2])

        p.add_pkg(x5)
        check_dbapi([x2])

    def test_installed(self):
        p = PackageTracker()
        x1 = self.make_pkg("/", "=dev-libs/X-1:0")
        x1b = self.make_pkg("/", "=dev-libs/X-1.1:0")
        x2 = self.make_pkg("/", "=dev-libs/X-2:0")
        x3 = self.make_pkg("/", "=dev-libs/X-3:1")

        def check_installed(x, should_contain, num_pkgs):
            self.assertEqual(x in p, should_contain)
            self.assertEqual(p.contains(x), should_contain)
            self.assertEqual(p.contains(x1, installed=True), should_contain)
            self.assertEqual(p.contains(x1, installed=False), False)
            self.assertEqual(len(list(p.all_pkgs("/"))), num_pkgs)

        def check_matches(atom, expected):
            matches = list(p.match("/", Atom(atom)))
            self.assertEqual(len(matches), len(expected))
            for x, y in zip(matches, expected):
                self.assertTrue(x is y)

        p.add_installed_pkg(x1)
        check_installed(x1, True, 1)
        check_matches("dev-libs/X", [x1])

        p.add_installed_pkg(x1)
        check_installed(x1, True, 1)
        check_matches("dev-libs/X", [x1])

        p.add_pkg(x2)
        check_installed(x1, False, 1)
        check_matches("dev-libs/X", [x2])

        p.add_installed_pkg(x1)
        check_installed(x1, False, 1)
        check_matches("dev-libs/X", [x2])

        p.add_installed_pkg(x1b)
        check_installed(x1, False, 1)
        check_installed(x1b, False, 1)
        check_matches("dev-libs/X", [x2])

        p.remove_pkg(x2)
        check_installed(x1, True, 2)
        check_installed(x1b, True, 2)
        check_matches("dev-libs/X", [x1, x1b])

    def test_conflicts(self):
        p = PackageTracker()
        installed1 = self.make_pkg("/", "=dev-libs/X-0:0")
        installed2 = self.make_pkg("/", "=dev-libs/X-0.1:0")
        x1 = self.make_pkg("/", "=dev-libs/X-1:0")
        x2 = self.make_pkg("/", "=dev-libs/X-2:0")
        x3 = self.make_pkg("/", "=dev-libs/X-3:0")
        x4 = self.make_pkg("/", "=dev-libs/X-4:4")
        x4b = self.make_pkg("/", "=dev-libs/X-4:4b::x-repo")

        def check_conflicts(expected, slot_conflicts_only=False):
            if slot_conflicts_only:
                conflicts = list(p.slot_conflicts())
            else:
                conflicts = list(p.conflicts())
            self.assertEqual(len(conflicts), len(expected))
            for got, exp in zip(conflicts, expected):
                self.assertEqual(got.description, exp.description)
                self.assertEqual(got.root, exp.root)
                self.assertEqual(len(got.pkgs), len(exp.pkgs))
                self.assertEqual(len(got), len(exp.pkgs))
                for x, y in zip(got.pkgs, exp.pkgs):
                    self.assertTrue(x is y)
                for x, y in zip(got, exp.pkgs):
                    self.assertTrue(x is y)
                for x in exp.pkgs:
                    self.assertTrue(x in got)

        check_conflicts([])
        check_conflicts([])

        p.add_installed_pkg(installed1)
        p.add_installed_pkg(installed2)
        check_conflicts([])

        p.add_pkg(x1)
        check_conflicts([])
        p.add_pkg(x2)
        check_conflicts([self.make_conflict("slot conflict", "/", [x1, x2])])
        p.add_pkg(x3)
        check_conflicts([self.make_conflict("slot conflict", "/", [x1, x2, x3])])
        p.remove_pkg(x3)
        check_conflicts([self.make_conflict("slot conflict", "/", [x1, x2])])
        p.remove_pkg(x2)
        check_conflicts([])
        p.add_pkg(x3)
        check_conflicts([self.make_conflict("slot conflict", "/", [x1, x3])])
        p.add_pkg(x2)
        check_conflicts([self.make_conflict("slot conflict", "/", [x1, x3, x2])])

        p.add_pkg(x4)
        check_conflicts([self.make_conflict("slot conflict", "/", [x1, x3, x2])])

        p.add_pkg(x4b)
        check_conflicts(
            [
                self.make_conflict("slot conflict", "/", [x1, x3, x2]),
                self.make_conflict("cpv conflict", "/", [x4, x4b]),
            ]
        )

        check_conflicts(
            [
                self.make_conflict("slot conflict", "/", [x1, x3, x2]),
            ],
            slot_conflicts_only=True,
        )



# Copyright 2014 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from __future__ import print_function
import sys

from portage.const import SUPPORTED_GENTOO_BINPKG_FORMATS
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)
from portage.output import colorize


class ChangedDepsTestCase(TestCase):
    def testChangedDeps(self):

        ebuilds = {
            "app-misc/A-0": {
                "DEPEND": "app-misc/B",
                "RDEPEND": "app-misc/B",
            },
            "app-misc/B-0": {},
        }

        binpkgs = {
            "app-misc/A-0": {},
        }

        installed = {
            "app-misc/A-0": {},
        }

        world = ("app-misc/A",)

        test_cases = (
            # --dynamic-deps=n causes the original deps to be respected
            ResolverPlaygroundTestCase(
                ["@world"],
                success=True,
                options={
                    "--update": True,
                    "--deep": True,
                    "--dynamic-deps": "n",
                    "--usepkg": True,
                },
                mergelist=[],
            ),
            # --dynamic-deps causes app-misc/B to get pulled in
            ResolverPlaygroundTestCase(
                ["@world"],
                success=True,
                options={
                    "--update": True,
                    "--deep": True,
                    "--dynamic-deps": "y",
                    "--usepkg": True,
                },
                mergelist=["app-misc/B-0"],
            ),
            # --changed-deps causes app-misc/A to be rebuilt
            ResolverPlaygroundTestCase(
                ["@world"],
                success=True,
                options={
                    "--update": True,
                    "--deep": True,
                    "--changed-deps": "y",
                    "--usepkg": True,
                },
                mergelist=["app-misc/B-0", "app-misc/A-0"],
            ),
            # --usepkgonly prevents automatic --binpkg-changed-deps
            ResolverPlaygroundTestCase(
                ["app-misc/A"],
                success=True,
                options={
                    "--changed-deps": "y",
                    "--usepkgonly": True,
                },
                mergelist=["[binary]app-misc/A-0"],
            ),
            # Test automatic --binpkg-changed-deps, which cases the
            # binpkg with stale deps to be ignored (with warning
            # message)
            ResolverPlaygroundTestCase(
                ["app-misc/A"],
                success=True,
                options={
                    "--usepkg": True,
                },
                mergelist=["app-misc/B-0", "app-misc/A-0"],
            ),
        )
        test_cases = (
            # Forcibly disable --binpkg-changed-deps, which causes
            # --changed-deps to be overridden by --binpkg-changed-deps
            ResolverPlaygroundTestCase(
                ["app-misc/A"],
                success=True,
                options={
                    "--binpkg-changed-deps": "n",
                    "--changed-deps": "y",
                    "--usepkg": True,
                },
                mergelist=["[binary]app-misc/A-0"],
            ),
        )

        for binpkg_format in SUPPORTED_GENTOO_BINPKG_FORMATS:
            with self.subTest(binpkg_format=binpkg_format):
                print(colorize("HILITE", binpkg_format), end=" ... ")
                sys.stdout.flush()
                playground = ResolverPlayground(
                    debug=False,
                    ebuilds=ebuilds,
                    binpkgs=binpkgs,
                    installed=installed,
                    world=world,
                    user_config={
                        "make.conf": ('BINPKG_FORMAT="%s"' % binpkg_format,),
                    },
                )

                try:
                    for test_case in test_cases:
                        playground.run_TestCase(test_case)
                        self.assertEqual(
                            test_case.test_success, True, test_case.fail_msg
                        )
                finally:
                    playground.cleanup()



# Copyright 2017 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from __future__ import print_function
import sys

from portage.const import SUPPORTED_GENTOO_BINPKG_FORMATS
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)
from portage.output import colorize


class BinaryPkgEbuildVisibilityTestCase(TestCase):
    def testBinaryPkgEbuildVisibility(self):

        binpkgs = {
            "app-misc/foo-3": {},
            "app-misc/foo-2": {},
            "app-misc/foo-1": {},
        }

        ebuilds = {
            "app-misc/foo-2": {},
            "app-misc/foo-1": {},
        }

        installed = {
            "app-misc/foo-1": {},
        }

        world = ["app-misc/foo"]

        test_cases = (
            # Test bug #612960, where --use-ebuild-visibility failed
            # to reject binary packages for which ebuilds were not
            # available.
            ResolverPlaygroundTestCase(
                ["@world"],
                options={
                    "--update": True,
                    "--deep": True,
                    "--use-ebuild-visibility": "y",
                    "--usepkgonly": True,
                },
                success=True,
                mergelist=[
                    "[binary]app-misc/foo-2",
                ],
            ),
            ResolverPlaygroundTestCase(
                ["@world"],
                options={
                    "--update": True,
                    "--deep": True,
                    "--usepkgonly": True,
                },
                success=True,
                mergelist=[
                    "[binary]app-misc/foo-3",
                ],
            ),
            ResolverPlaygroundTestCase(
                ["@world"],
                options={
                    "--update": True,
                    "--deep": True,
                    "--usepkg": True,
                },
                success=True,
                mergelist=[
                    "[binary]app-misc/foo-2",
                ],
            ),
            ResolverPlaygroundTestCase(
                ["=app-misc/foo-3"],
                options={
                    "--use-ebuild-visibility": "y",
                    "--usepkgonly": True,
                },
                success=False,
            ),
            ResolverPlaygroundTestCase(
                ["app-misc/foo"],
                options={
                    "--use-ebuild-visibility": "y",
                    "--usepkgonly": True,
                },
                success=True,
                mergelist=[
                    "[binary]app-misc/foo-2",
                ],
            ),
            ResolverPlaygroundTestCase(
                ["app-misc/foo"],
                options={
                    "--usepkgonly": True,
                },
                success=True,
                mergelist=[
                    "[binary]app-misc/foo-3",
                ],
            ),
            # The default behavior is to enforce ebuild visibility as
            # long as a visible package is available to satisfy the
            # current atom. In the following test case, ebuild visibility
            # is ignored in order to satisfy the =app-misc/foo-3 atom.
            ResolverPlaygroundTestCase(
                ["=app-misc/foo-3"],
                options={
                    "--usepkg": True,
                },
                success=True,
                mergelist=[
                    "[binary]app-misc/foo-3",
                ],
            ),
            # Verify that --use-ebuild-visibility works with --usepkg
            # when no other visible package is available.
            ResolverPlaygroundTestCase(
                ["=app-misc/foo-3"],
                options={
                    "--use-ebuild-visibility": "y",
                    "--usepkg": True,
                },
                success=False,
            ),
        )

        for binpkg_format in SUPPORTED_GENTOO_BINPKG_FORMATS:
            with self.subTest(binpkg_format=binpkg_format):
                print(colorize("HILITE", binpkg_format), end=" ... ")
                sys.stdout.flush()
                playground = ResolverPlayground(
                    binpkgs=binpkgs,
                    ebuilds=ebuilds,
                    installed=installed,
                    world=world,
                    user_config={
                        "make.conf": ('BINPKG_FORMAT="%s"' % binpkg_format,),
                    },
                )
                try:
                    for test_case in test_cases:
                        playground.run_TestCase(test_case)
                        self.assertEqual(
                            test_case.test_success, True, test_case.fail_msg
                        )
                finally:
                    playground.cleanup()



# elog/mod_mail_summary.py - elog dispatch module
# Copyright 2006-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import portage
from portage.exception import AlarmSignal, PortageException
from portage.localization import _
from portage.util import writemsg
from portage import _encodings
from portage import _unicode_decode

import socket
import time

_config_keys = (
    "PORTAGE_ELOG_MAILURI",
    "PORTAGE_ELOG_MAILFROM",
    "PORTAGE_ELOG_MAILSUBJECT",
)
_items = {}


def process(mysettings, key, logentries, fulltext):
    global _items
    time_str = _unicode_decode(
        time.strftime("%Y%m%d-%H%M%S %Z", time.localtime(time.time())),
        encoding=_encodings["content"],
        errors="replace",
    )
    header = _(
        ">>> Messages generated for package %(pkg)s by process %(pid)d on %(time)s:\n\n"
    ) % {"pkg": key, "pid": portage.getpid(), "time": time_str}
    config_root = mysettings["PORTAGE_CONFIGROOT"]

    # Copy needed variables from the config instance,
    # since we don't need to hold a reference for the
    # whole thing. This also makes it possible to
    # rely on per-package variable settings that may
    # have come from /etc/portage/package.env, since
    # we'll be isolated from any future mutations of
    # mysettings.
    config_dict = {}
    for k in _config_keys:
        v = mysettings.get(k)
        if v is not None:
            config_dict[k] = v

    config_dict, items = _items.setdefault(config_root, (config_dict, {}))
    items[key] = header + fulltext


def finalize():
    global _items
    for mysettings, items in _items.values():
        _finalize(mysettings, items)
    _items.clear()


def _finalize(mysettings, items):
    if len(items) == 0:
        return
    if len(items) == 1:
        count = _("one package")
    else:
        count = _("multiple packages")
    if "PORTAGE_ELOG_MAILURI" in mysettings:
        myrecipient = mysettings["PORTAGE_ELOG_MAILURI"].split()[0]
    else:
        myrecipient = "root@localhost"

    myfrom = mysettings.get("PORTAGE_ELOG_MAILFROM", "")
    myfrom = myfrom.replace("${HOST}", socket.getfqdn())
    mysubject = mysettings.get("PORTAGE_ELOG_MAILSUBJECT", "")
    mysubject = mysubject.replace("${PACKAGE}", count)
    mysubject = mysubject.replace("${HOST}", socket.getfqdn())

    mybody = _(
        "elog messages for the following packages generated by "
        "process %(pid)d on host %(host)s:\n"
    ) % {"pid": portage.getpid(), "host": socket.getfqdn()}
    for key in items:
        mybody += "- %s\n" % key

    mymessage = portage.mail.create_message(
        myfrom, myrecipient, mysubject, mybody, attachments=list(items.values())
    )

    # Timeout after one minute in case send_mail() blocks indefinitely.
    try:
        try:
            AlarmSignal.register(60)
            portage.mail.send_mail(mysettings, mymessage)
        finally:
            AlarmSignal.unregister()
    except AlarmSignal:
        writemsg(
            "Timeout in finalize() for elog system 'mail_summary'\n", noiselevel=-1
        )
    except PortageException as e:
        writemsg("%s\n" % (e,), noiselevel=-1)

    return



# Copyright 1999-2012, 2016 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import logging

import portage
from portage import os
from portage.localization import _
from portage.output import bold, colorize, yellow
from portage.util import writemsg_level


def chk_updated_cfg_files(eroot, config_protect):
    target_root = eroot
    result = list(portage.util.find_updated_config_files(target_root, config_protect))

    for x in result:
        writemsg_level(
            "\n %s " % (colorize("WARN", "* " + _("IMPORTANT:"))),
            level=logging.INFO,
            noiselevel=-1,
        )
        if not x[1]:  # it's a protected file
            writemsg_level(
                _("config file '%s' needs updating.\n") % x[0],
                level=logging.INFO,
                noiselevel=-1,
            )
        else:  # it's a protected dir
            if len(x[1]) == 1:
                head, tail = os.path.split(x[1][0])
                tail = tail[len("._cfg0000_") :]
                fpath = os.path.join(head, tail)
                writemsg_level(
                    _("config file '%s' needs updating.\n") % fpath,
                    level=logging.INFO,
                    noiselevel=-1,
                )
            else:
                writemsg_level(
                    _("%d config files in '%s' need updating.\n") % (len(x[1]), x[0]),
                    level=logging.INFO,
                    noiselevel=-1,
                )

    if result:
        print(
            " "
            + yellow("*")
            + " See the "
            + colorize("INFORM", _("CONFIGURATION FILES"))
            + " and "
            + colorize("INFORM", _("CONFIGURATION FILES UPDATE TOOLS"))
        )
        print(
            " "
            + yellow("*")
            + " sections of the "
            + bold("emerge")
            + " "
            + _("man page to learn how to update config files.")
        )



# Copyright 2018 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import functools

from portage.util._async.AsyncTaskFuture import AsyncTaskFuture
from portage.util._async.TaskScheduler import TaskScheduler
from portage.util.futures import asyncio
from portage.util.cpuinfo import get_cpu_count


def iter_completed(futures, max_jobs=None, max_load=None, loop=None):
    """
    This is similar to asyncio.as_completed, but takes an iterator of
    futures as input, and includes support for max_jobs and max_load
    parameters.

    @param futures: iterator of asyncio.Future (or compatible)
    @type futures: iterator
    @param max_jobs: max number of futures to process concurrently (default
            is portage.util.cpuinfo.get_cpu_count())
    @type max_jobs: int
    @param max_load: max load allowed when scheduling a new future,
            otherwise schedule no more than 1 future at a time (default
            is portage.util.cpuinfo.get_cpu_count())
    @type max_load: int or float
    @param loop: event loop
    @type loop: EventLoop
    @return: iterator of futures that are done
    @rtype: iterator
    """
    loop = asyncio._wrap_loop(loop)

    for future_done_set in async_iter_completed(
        futures, max_jobs=max_jobs, max_load=max_load, loop=loop
    ):
        for future in loop.run_until_complete(future_done_set):
            yield future


def async_iter_completed(futures, max_jobs=None, max_load=None, loop=None):
    """
    An asynchronous version of iter_completed. This yields futures, which
    when done, result in a set of input futures that are done. This serves
    as a wrapper around portage's internal TaskScheduler class, using
    standard asyncio interfaces.

    @param futures: iterator of asyncio.Future (or compatible)
    @type futures: iterator
    @param max_jobs: max number of futures to process concurrently (default
            is portage.util.cpuinfo.get_cpu_count())
    @type max_jobs: int
    @param max_load: max load allowed when scheduling a new future,
            otherwise schedule no more than 1 future at a time (default
            is portage.util.cpuinfo.get_cpu_count())
    @type max_load: int or float
    @param loop: event loop
    @type loop: EventLoop
    @return: iterator of futures, which when done, result in a set of
            input futures that are done
    @rtype: iterator
    """
    loop = asyncio._wrap_loop(loop)

    max_jobs = max_jobs or get_cpu_count()
    max_load = max_load or get_cpu_count()

    future_map = {}

    def task_generator():
        for future in futures:
            future_map[id(future)] = future
            yield AsyncTaskFuture(future=future)

    scheduler = TaskScheduler(
        task_generator(), max_jobs=max_jobs, max_load=max_load, event_loop=loop
    )

    def done_callback(future_done_set, wait_result):
        """Propagate results from wait_result to future_done_set."""
        if future_done_set.cancelled():
            return
        done, pending = wait_result.result()
        for future in done:
            del future_map[id(future)]
        future_done_set.set_result(done)

    def cancel_callback(wait_result, future_done_set):
        """Cancel wait_result if future_done_set has been cancelled."""
        if future_done_set.cancelled() and not wait_result.done():
            wait_result.cancel()

    try:
        scheduler.start()

        # scheduler should ensure that future_map is non-empty until
        # task_generator is exhausted
        while future_map:
            wait_result = asyncio.ensure_future(
                asyncio.wait(
                    list(future_map.values()),
                    return_when=asyncio.FIRST_COMPLETED,
                    loop=loop,
                ),
                loop=loop,
            )
            future_done_set = loop.create_future()
            future_done_set.add_done_callback(
                functools.partial(cancel_callback, wait_result)
            )
            wait_result.add_done_callback(
                functools.partial(done_callback, future_done_set)
            )
            yield future_done_set
    finally:
        # cleanup in case of interruption by SIGINT, etc
        if not loop.is_closed():
            scheduler.cancel()


def iter_gather(futures, max_jobs=None, max_load=None, loop=None):
    """
    This is similar to asyncio.gather, but takes an iterator of
    futures as input, and includes support for max_jobs and max_load
    parameters.

    @param futures: iterator of asyncio.Future (or compatible)
    @type futures: iterator
    @param max_jobs: max number of futures to process concurrently (default
            is portage.util.cpuinfo.get_cpu_count())
    @type max_jobs: int
    @param max_load: max load allowed when scheduling a new future,
            otherwise schedule no more than 1 future at a time (default
            is portage.util.cpuinfo.get_cpu_count())
    @type max_load: int or float
    @param loop: event loop
    @type loop: EventLoop
    @return: a Future resulting in a list of done input futures, in the
            same order that they were yielded from the input iterator
    @rtype: asyncio.Future (or compatible)
    """
    loop = asyncio._wrap_loop(loop)
    result = loop.create_future()
    futures_list = []

    def future_generator():
        for future in futures:
            futures_list.append(future)
            yield future

    completed_iter = async_iter_completed(
        future_generator(),
        max_jobs=max_jobs,
        max_load=max_load,
        loop=loop,
    )

    def handle_result(future_done_set):
        if result.cancelled():
            if not future_done_set.cancelled():
                # All exceptions must be consumed from future_done_set, in order
                # to avoid triggering the event loop's exception handler.
                list(
                    future.exception()
                    for future in future_done_set.result()
                    if not future.cancelled()
                )
            return

        try:
            handle_result.current_task = next(completed_iter)
        except StopIteration:
            result.set_result(futures_list)
        else:
            handle_result.current_task.add_done_callback(handle_result)

    try:
        handle_result.current_task = next(completed_iter)
    except StopIteration:
        handle_result.current_task = None
        result.set_result(futures_list)
    else:
        handle_result.current_task.add_done_callback(handle_result)

    def cancel_callback(result):
        if (
            result.cancelled()
            and handle_result.current_task is not None
            and not handle_result.current_task.done()
        ):
            handle_result.current_task.cancel()

    result.add_done_callback(cancel_callback)

    return result



# Copyright 1999-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import logging
import textwrap

import portage
from portage import os
from portage.emaint.modules.logs.logs import CleanLogs
from portage.news import count_unread_news, display_news_notifications
from portage.output import colorize
from portage.util._dyn_libs.display_preserved_libs import display_preserved_libs
from portage.util._info_files import chk_updated_info_files

from .chk_updated_cfg_files import chk_updated_cfg_files
from .emergelog import emergelog
from ._flush_elog_mod_echo import _flush_elog_mod_echo


def clean_logs(settings):

    if "clean-logs" not in settings.features:
        return

    logdir = settings.get("PORTAGE_LOGDIR")
    if logdir is None or not os.path.isdir(logdir):
        return

    cleanlogs = CleanLogs()
    returncode, msgs = cleanlogs.clean(settings=settings)
    if not returncode:
        out = portage.output.EOutput()
        for msg in msgs:
            out.eerror(msg)


def display_news_notification(root_config, myopts):
    if "news" not in root_config.settings.features:
        return False
    portdb = root_config.trees["porttree"].dbapi
    vardb = root_config.trees["vartree"].dbapi
    news_counts = count_unread_news(portdb, vardb)
    if all(v == 0 for v in news_counts.values()):
        return False
    display_news_notifications(news_counts)
    return True


def show_depclean_suggestion():
    out = portage.output.EOutput()
    msg = (
        "After world updates, it is important to remove "
        + "obsolete packages with emerge --depclean. Refer "
        + "to `man emerge` for more information."
    )
    for line in textwrap.wrap(msg, 72):
        out.ewarn(line)


def post_emerge(myaction, myopts, myfiles, target_root, trees, mtimedb, retval):
    """
    Misc. things to run at the end of a merge session.

    Update Info Files
    Update Config Files
    Update News Items
    Commit mtimeDB
    Display preserved libs warnings

    @param myaction: The action returned from parse_opts()
    @type myaction: String
    @param myopts: emerge options
    @type myopts: dict
    @param myfiles: emerge arguments
    @type myfiles: list
    @param target_root: The target EROOT for myaction
    @type target_root: String
    @param trees: A dictionary mapping each ROOT to it's package databases
    @type trees: dict
    @param mtimedb: The mtimeDB to store data needed across merge invocations
    @type mtimedb: MtimeDB class instance
    @param retval: Emerge's return value
    @type retval: Int
    """

    root_config = trees[target_root]["root_config"]
    vardbapi = trees[target_root]["vartree"].dbapi
    settings = vardbapi.settings
    info_mtimes = mtimedb["info"]

    # Load the most current variables from ${ROOT}/etc/profile.env
    settings.unlock()
    settings.reload()
    settings.regenerate()
    settings.lock()

    config_protect = portage.util.shlex_split(settings.get("CONFIG_PROTECT", ""))
    infodirs = settings.get("INFOPATH", "").split(":") + settings.get(
        "INFODIR", ""
    ).split(":")

    os.chdir("/")

    if retval == os.EX_OK:
        exit_msg = " *** exiting successfully."
    else:
        exit_msg = " *** exiting unsuccessfully with status '%s'." % retval
    emergelog("notitles" not in settings.features, exit_msg)

    _flush_elog_mod_echo()

    if not vardbapi._pkgs_changed:
        # GLEP 42 says to display news *after* an emerge --pretend
        if "--pretend" in myopts:
            display_news_notification(root_config, myopts)
        # If vdb state has not changed then there's nothing else to do.
        return

    vdb_path = os.path.join(root_config.settings["EROOT"], portage.VDB_PATH)
    portage.util.ensure_dirs(vdb_path)
    vdb_lock = None
    if os.access(vdb_path, os.W_OK) and not "--pretend" in myopts:
        vardbapi.lock()
        vdb_lock = True

    if vdb_lock:
        try:
            if "noinfo" not in settings.features:
                chk_updated_info_files(target_root, infodirs, info_mtimes)
            mtimedb.commit()
        finally:
            if vdb_lock:
                vardbapi.unlock()

    # Explicitly load and prune the PreservedLibsRegistry in order
    # to ensure that we do not display stale data.
    vardbapi._plib_registry.load()

    if vardbapi._plib_registry.hasEntries():
        if "--quiet" in myopts:
            print()
            print(colorize("WARN", "!!!") + " existing preserved libs found")
        else:
            print()
            print(colorize("WARN", "!!!") + " existing preserved libs:")
            display_preserved_libs(vardbapi, verbose="--verbose" in myopts)
            print(
                "Use "
                + colorize("GOOD", "emerge @preserved-rebuild")
                + " to rebuild packages using these libraries"
            )

    chk_updated_cfg_files(settings["EROOT"], config_protect)

    display_news_notification(root_config, myopts)

    postemerge = os.path.join(
        settings["PORTAGE_CONFIGROOT"], portage.USER_CONFIG_PATH, "bin", "post_emerge"
    )
    if os.access(postemerge, os.X_OK):
        hook_retval = portage.process.spawn([postemerge], env=settings.environ())
        if hook_retval != os.EX_OK:
            portage.util.writemsg_level(
                " %s spawn failed of %s\n"
                % (
                    colorize("BAD", "*"),
                    postemerge,
                ),
                level=logging.ERROR,
                noiselevel=-1,
            )

    clean_logs(settings)

    if "--quiet" not in myopts and myaction is None and "@world" in myfiles:
        show_depclean_suggestion()



# Copyright 2018 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import time
from portage.tests import TestCase
from portage.util._async.ForkProcess import ForkProcess
from portage.util._eventloop.global_event_loop import global_event_loop
from portage.util.futures import asyncio
from portage.util.futures.iter_completed import (
    iter_completed,
    async_iter_completed,
)


class SleepProcess(ForkProcess):
    __slots__ = ("future", "seconds")

    def _start(self):
        self.addExitListener(self._future_done)
        ForkProcess._start(self)

    def _future_done(self, task):
        if not self.future.cancelled():
            self.future.set_result(self.seconds)

    def _run(self):
        time.sleep(self.seconds)


class IterCompletedTestCase(TestCase):
    def testIterCompleted(self):

        # Mark this as todo, since we don't want to fail if heavy system
        # load causes the tasks to finish in an unexpected order.
        self.todo = True

        loop = global_event_loop()
        tasks = [
            SleepProcess(seconds=0.200),
            SleepProcess(seconds=0.100),
            SleepProcess(seconds=0.001),
        ]

        expected_order = sorted(task.seconds for task in tasks)

        def future_generator():
            for task in tasks:
                task.future = loop.create_future()
                task.scheduler = loop
                task.start()
                yield task.future

        for seconds, future in zip(
            expected_order,
            iter_completed(future_generator(), max_jobs=True, max_load=None, loop=loop),
        ):
            self.assertEqual(seconds, future.result())

    def testAsyncCancel(self):

        loop = global_event_loop()
        input_futures = set()
        future_count = 3

        def future_generator():
            for i in range(future_count):
                future = loop.create_future()
                loop.call_soon(
                    lambda future: None if future.done() else future.set_result(None),
                    future,
                )
                input_futures.add(future)
                yield future

        for future_done_set in async_iter_completed(
            future_generator(), max_jobs=True, max_load=True, loop=loop
        ):
            future_done_set.cancel()
            break

        # With max_jobs=True, async_iter_completed should have executed
        # the generator until it raised StopIteration.
        self.assertEqual(future_count, len(input_futures))

        loop.run_until_complete(asyncio.wait(input_futures, loop=loop))

        # The futures may have results or they may have been cancelled
        # by TaskScheduler, and behavior varies depending on the python
        # interpreter.
        for future in input_futures:
            future.cancelled() or future.result()



# Copyright 1999-2014 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage import os
from portage.dep import _repo_separator
from portage.output import colorize

from _emerge.AsynchronousTask import AsynchronousTask
from _emerge.Binpkg import Binpkg
from _emerge.CompositeTask import CompositeTask
from _emerge.EbuildBuild import EbuildBuild
from _emerge.PackageUninstall import PackageUninstall


class MergeListItem(CompositeTask):

    """
    TODO: For parallel scheduling, everything here needs asynchronous
    execution support (start, poll, and wait methods).
    """

    __slots__ = (
        "args_set",
        "binpkg_opts",
        "build_opts",
        "config_pool",
        "emerge_opts",
        "find_blockers",
        "logger",
        "mtimedb",
        "pkg",
        "pkg_count",
        "pkg_to_replace",
        "prefetcher",
        "settings",
        "statusMessage",
        "world_atom",
    ) + ("_install_task",)

    def _start(self):

        pkg = self.pkg
        build_opts = self.build_opts

        if pkg.installed:
            # uninstall,  executed by self.merge()
            self.returncode = os.EX_OK
            self._async_wait()
            return

        args_set = self.args_set
        find_blockers = self.find_blockers
        logger = self.logger
        mtimedb = self.mtimedb
        pkg_count = self.pkg_count
        scheduler = self.scheduler
        settings = self.settings
        world_atom = self.world_atom
        ldpath_mtimes = mtimedb["ldpath"]

        action_desc = "Emerging"
        preposition = "for"
        pkg_color = "PKG_MERGE"
        if pkg.type_name == "binary":
            pkg_color = "PKG_BINARY_MERGE"
            action_desc += " binary"

        if build_opts.fetchonly:
            action_desc = "Fetching"

        msg = "%s (%s of %s) %s" % (
            action_desc,
            colorize("MERGE_LIST_PROGRESS", str(pkg_count.curval)),
            colorize("MERGE_LIST_PROGRESS", str(pkg_count.maxval)),
            colorize(pkg_color, pkg.cpv + _repo_separator + pkg.repo),
        )

        if pkg.root_config.settings["ROOT"] != "/":
            msg += " %s %s" % (preposition, pkg.root)

        if not build_opts.pretend:
            self.statusMessage(msg)
            logger.log(
                " >>> emerge (%s of %s) %s to %s"
                % (pkg_count.curval, pkg_count.maxval, pkg.cpv, pkg.root)
            )

        if pkg.type_name == "ebuild":

            build = EbuildBuild(
                args_set=args_set,
                background=self.background,
                config_pool=self.config_pool,
                find_blockers=find_blockers,
                ldpath_mtimes=ldpath_mtimes,
                logger=logger,
                opts=build_opts,
                pkg=pkg,
                pkg_count=pkg_count,
                prefetcher=self.prefetcher,
                scheduler=scheduler,
                settings=settings,
                world_atom=world_atom,
            )

            self._install_task = build
            self._start_task(build, self._default_final_exit)
            return

        if pkg.type_name == "binary":

            binpkg = Binpkg(
                background=self.background,
                find_blockers=find_blockers,
                ldpath_mtimes=ldpath_mtimes,
                logger=logger,
                opts=self.binpkg_opts,
                pkg=pkg,
                pkg_count=pkg_count,
                prefetcher=self.prefetcher,
                settings=settings,
                scheduler=scheduler,
                world_atom=world_atom,
            )

            self._install_task = binpkg
            self._start_task(binpkg, self._default_final_exit)
            return

    def create_install_task(self):

        pkg = self.pkg
        build_opts = self.build_opts
        mtimedb = self.mtimedb
        scheduler = self.scheduler
        settings = self.settings
        world_atom = self.world_atom
        ldpath_mtimes = mtimedb["ldpath"]

        if pkg.installed:
            if not (
                build_opts.buildpkgonly or build_opts.fetchonly or build_opts.pretend
            ):

                task = PackageUninstall(
                    background=self.background,
                    ldpath_mtimes=ldpath_mtimes,
                    opts=self.emerge_opts,
                    pkg=pkg,
                    scheduler=scheduler,
                    settings=settings,
                    world_atom=world_atom,
                )

            else:
                task = AsynchronousTask()

        elif build_opts.fetchonly or build_opts.buildpkgonly:
            task = AsynchronousTask()
        else:
            task = self._install_task.create_install_task()

        return task
